import threading,traceback,math
from typing import List
from time import time,sleep

def get_tab_from_locals(loc):
	return {k : v for k,v in loc.items() if k != 'self'}


class ThreadOptions:
	time_delta : int = None
	"""
	dynamic_time : next run will be at (interval/sleep_time - runtime)
	"""
	def __init__(self,
			interval        : int  = None,
			start_with_wait        = False,
			dynamic_time           = False,
			condition              = None,
			daemon                 = True,
			sleep_time      : int  = None,
			no_loop         : bool = False,
			debug           : bool = False,
			first_run:callable=None):

		self.set(**get_tab_from_locals(locals()))
	"""
	dynamic_time : next run will be at (interval/sleep_time - runtime)
	"""
	def set(self,
			interval        : int  = None,
			start_with_wait        = False,
			dynamic_time           = False,
			condition              = None,
			daemon                 = True,
			sleep_time      : int  = None,
			no_loop         : bool = False,
			debug           : bool = False,
			first_run:callable=None):
		
		for k,v in get_tab_from_locals(locals()).items():
			setattr(self,k,v)

		
		return self
	



	

class ThreadOptionsHelper_old:
	def __init__(self,ops:ThreadOptions):
		self.interval	     = ops.interval
		self.start_with_wait = ops.start_with_wait
		self.dynamic_time    = ops.dynamic_time
		self.condition		 = ops.condition
		self.daemon			 = ops.daemon
		self.sleep_time		 = ops.sleep_time
		self.time_delta		 = self.interval if self.interval is not None else (self.sleep_time if self.sleep_time is not None else None)



	def set_start_time(self):
		self._start_time = time()
	
	def sleep_first_time(self):
		if self.time_delta is None:
			return None
	def sleep(self):
		if not self.time_delta:
			return
		
		stop_time = time()
		time_to_sleep = max(0,self.time_delta - ( stop_time - self._start_time )) if self.dynamic_time else self.time_delta
		self._start_time = stop_time
		if not time_to_sleep:
			return
		sleep(time_to_sleep)

	

class ThreadOptionsHelper:
	def __init__(self,ops:ThreadOptions,func=None,args=[],kwargs=dict()):
		self.ops 			 = ops
		self.time_delta		 = self.ops.interval if self.ops.interval is not None else (self.ops.sleep_time if self.ops.sleep_time is not None else None)
		self.func 			 = func
		self.args 			 = args
		self.kwargs 		 = kwargs
	
		self.start()
	
	def info(self,*a):
		if not self.ops.debug:
			return
		print(f'Thread {self.func.__name__}',*a)
	def start(self):

		def thread_function(*a,**b):

			t = threading.currentThread()
			def condition_func(*a,**b):
				if not getattr(t,'is_running',True):
					print('stopped running!')
					return False
				if self.ops.condition:
					return self.ops.condition(*a,**b)
				return True
		
			if self.ops.first_run:
				self.ops.first_run(*a,**b)
			self.sleep_first_time()
			self._start_time = time()
			while condition_func(*a,**b):
			
				self.func(*a,**b)
				self.sleep()
				if self.ops.no_loop:
					break
				self._start_time = time()
		
		t = threading.Thread(target=thread_function,args=self.args,kwargs=self.kwargs,daemon=self.ops.daemon)
		setattr(t,"is_running",True)
		t.start()

					
	
	def sleep_first_time(self):
		if self.time_delta is None:
			return None
		if self.ops.start_with_wait:
			sleep(self.time_delta)
	def sleep(self):
		if not self.time_delta:
			return
		
		
		stop_time = time()
	
		time_to_sleep = max(0,self.time_delta - ( stop_time - self._start_time )) if self.ops.dynamic_time else self.time_delta

		if not time_to_sleep:

			return
		sleep(time_to_sleep)
	
class ThreadClass:

	@property
	def options(self):
		return ThreadOptions().set

	@staticmethod
	def conditionProp(ob,propname,eq=True,neq=None):
		def ret_func(*a,**b):
			val = getattr(ob,propname)
			if eq is not None:
				return val == eq
			if neq is not None:
				return val != neq
			return False
		return ret_func
	
	def register(self,options:ThreadOptions=ThreadOptions(),func=None,args=[],kwargs=dict()):
		if func:
			return ThreadOptionsHelper(options,func=func,args=args,kwargs=kwargs)
		
		def add_func(f):
			ThreadOptionsHelper(options,func=f,args=args,kwargs=kwargs)
			return f
		
		return add_func
	
	def register_old(self,options:ThreadOptions=ThreadOptions(),func=None,args=[],kwargs=dict()):
		helper = ThreadOptionsHelper(options)
		def add_func(f):
			def func(*a,**b):
				t = threading.currentThread()
				def condition_func(*a,**b):
					if not getattr(t,'is_running',True):
						print('stopped running!')
						return False
					if options.condition:
						return options.condition(*a,**b)
					return True
				
				helper.sleep_first_time()
				helper.set_start_time()
				while condition_func(*a,**b):
					
					f(*a,**b)
					
					helper.sleep()
					if options.no_loop:
						break
				
			t = threading.Thread(target=func,args=args,kwargs=kwargs,daemon=options.daemon)
			setattr(t,"is_running",True)
			t.start()
			
			return f
		
		if func:
			return add_func(func)
		
		return add_func

	@classmethod
	def stop_all(cls,max_wait_for_it:int=0):
		print(f'stop all started {len(threading.enumerate())}')
		
		for t in threading.enumerate():
			try:
				setattr(t,'is_running',False)
				t.join(0.1)
			except:
				pass
		
		if max_wait_for_it:
			max_time   = time() + max_wait_for_it
			while time() < max_time:
				if not threading.activeCount():
					return True
				sleep(0.1)
			print(threading.activeCount())
			return False
		else:
			return not threading.activeCount()
		

				
					



Thread = ThreadClass()





from __future__ 			import annotations
from .base import BaseView as B
from typing import List
from ...tcpsock import TCPClient,TCPServer
import json
from ...event import EventWithKey
from ...Curses import valid_key,get_keyname

server_port = 4124
client_port = 4123

class Tetris(B.Tetris):
	def init(self):
		self.nextBlock    = B.NextBlock()
		self.holdingBlock = B.HoldingBlock()
		self.box          = B.Box()
		self.score        = B.Score()
		self.on_key       = EventWithKey(valid_key=valid_key,format_key=get_keyname)
		self._register_ev()
	def on_receive(self,data:dict):
		if key_press := data.get("key",None):
			self.on_key.fire(key_press)
			return
		
	
	def _register_ev(self):
		ok = self.on_key.register
		mapper = dict(
			esc=self.ctrlTrg.do_exit,
			left=self.ctrlTrg.left,
			right=self.ctrlTrg.right,
			down=self.ctrlTrg.down,
			up=self.ctrlTrg.up,
			dc=self.ctrlTrg.holding,
			space=self.ctrlTrg.drop)
			
		for k,v in mapper.items():
			self.on_key.register(k,self.on_key.cb(clear_args=True),func=v.fire)


	def start(self):
		self.server = TCPServer(port=server_port,cb=self.on_receive,decode="utf8",is_json=True)
		
	def send(self,msg):
		if isinstance(msg,dict):
			msg = json.dumps(msg)
		TCPClient.send(client_port,msg=msg)
	
	def check_changes(self):
		has_changed = False
		for k,v in self.elements.items():
			if self.dispData.lastChanged[k] > v.last_update:
				v.last_update = self.dispData.lastChanged[k]
				has_changed = True
			
		if has_changed:
			self.send({k : v().to_dict() for k,v in self.dispData.getData.items()})


	def exit(self):
		self.send(dict(electron_command="exit"))
		self.running = False
		self.server.shutdown()
		


from .tester import Tester
from ..menu.plugin import *

class PluginTester(Tester):
	def test(self):
		def get_cb(name):
			def cb(*a,**b):
				print(name,*a,**b)

			return cb

		def explorer():
			e = Explorer("Explorer test",cb=get_cb("explor"))
			e.display()
		def selecter():
			s = Selecter("Selecter test",SelecterOption("option a",1),SelecterOption("option b",2),cb=get_cb("se"))
		def yon():
			i = YesNoQuestion("Werkt dit ?",cb=get_cb("yon"))
		def textinput():
			i = TextInput("TextInput test",cb=get_cb("text input"))

		return [explorer,selecter,yon,textinput]


from ..models import string_replacer,InputParameter,copy_tab
from . import UIInput
from typing import List

class Item:

	_title_base : str            = "Getvaluefor[name]"
	_desc_base  : str            = "Getvaluefor[name]"
	para        : InputParameter
	uiInput     : UIInput          
	def __init__(self,uiInput:UIInput):
		self.uiInput = uiInput


	@property
	def para(self):
		return self.uiInput.para
	
	@property
	def desc(self):
		return string_replacer(self._desc_base,self.replacement_tab)

	@property
	def title(self):
		return string_replacer(self._title_base,self.replacement_tab)
	
	@property
	def my_replacement_tab(self):
		return dict()
	
	@property
	def replacement_tab(self):
		tab = copy_tab(self.uiInput.replacement_tab)
		tab.update(self.my_replacement_tab)
		return tab

	def __repr__(self):
		return f"[<{self.__class__.__name__}>] {self.title}"

class FileSelecter(Item):

	_desc_base  = "Browse for [name]"
	_title_base = "Please select file [name]"

class String(Item):
	_desc_base  = "Input text for [name]"
	_title_base = "Please input [name]"


class Number(String):
	_desc_base  = "Input number for [name]"
	_title_base = "Please input number [name]"


class Filepath(String):
	_desc_base  = "Input path to [name]"
	_title_base = "Please the filepath to [name]"


class Text(Item):
	_desc_base  = "Input text for [name]"
	_title_base = "Please input [name]"


import math,os,subprocess
import random 								as _random
from datetime 		import 	datetime
from typing 		import 	List
from urllib.request import 	Request, urlopen
from _thread 		import 	start_new_thread as Thread

random_el = _random.choice


def index_random(ar):
	return _random.randint(0,len(ar) - 1)
def enumerate_random(ar):
	i_random = index_random(ar)
	return i_random,ar[i_random]

def enumerate_reversed(ar):
	return [[i,ar[i]] for i in indexes_reversed(ar)]



def indexes_reversed(ar):

	return [i for i in range(len(ar) -1, -1, -1)]


def indexes(ar):

	return [i for i in range(len(ar) -1, -1, -1)]


def fill_to_len(s:str,min_len:int,filler=' ',left_side=True):
	if len(s) >= min_len:
		return s
	if left_side:
		return filler*(min_len - len(s)) + s
	return s + filler*(min_len - len(s))

def between(s:str,a:str,b:str) -> str:
	try:
		x = s.find(a)
		y = s.find(b,x+len(a))
		retval = s[x+len(a):y]
		return retval
	except:
		return None

def random(n=10):
	letters = 'abcdefghijklmnopqrstuvwxyz0123456789'
	return ''.join(_random.choice(letters) for i in range(n))

def same_len_ar(ar,to_string=True):
	max_lens = []
	for i,a in enumerate(ar):
		if not i:
			max_lens = [len(a_el) for a_el in a]
		else:
			max_lens = [max(max_lens[a_ind],len(a[a_ind])) for a_ind in range(len(a))]
	
	ret_val = []
	for x in ar:
		ret_val.append(' '.join([fill_to_len(s,min_len=max_lens[index],left_side=False) for index,s in enumerate(x)]))
	return ret_val

def copy_tab(t):
	return {k : v for k,v in t.items()}

def properties_keys_tab(ob):
	return {k : True for k in ob.__dict__.keys()}

def properties_keys_ar(ob):
	return list(ob.__dict__.keys())

def properties_set_all(ob,proptabtoset=None,**t):
	if t:
		proptabtoset = t
	
	if proptabtoset:
		for k,v in proptabtoset.items():
			setattr(ob,k,v)


def is_str_list(val):
	if not isinstance(val,list):
		return False

	for x in val:
		if not isinstance(x,str):
			return False
	return True


'''
class A:
	vv = ['a']

class B(A):
	vv = ['b']

class C(B):
	vv = ['c']


class D(C):
	vv = ['d']

=> ['d','c','b','a']

'''
def get_value_stack_from_supers(ob,propname,unique=False):
	retval = getattr(ob,propname,[])
	for c in ob.__class__.__mro__:
		x = super(c,ob)
		if hasattr(x,propname):
			retval += getattr(x,propname)
		else:
			break
	
	return set(retval) if unique else retval
		
def download_html(url,get_body=True):
	def getBody(txt):
		body = txt
		try:
			tl = txt.lower()
			a = tl.find('<body')
			b = tl.find('>',a+1)
			c = tl.find('</body>',b)
			body = txt[b+1:c]
		except:
			pass
		return body
	user_agent = 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2272.101 Safari/537.36'
	headers = { 'User-Agent' : user_agent }
	req = Request(url, None, headers)
	resp = urlopen(req)

	html = resp.read()
	try:
		txt = html.decode()
		return getBody(txt) if get_body else txt
	except:
		return html

def Popen(s):
	try:
		process = subprocess.Popen(s, stdout=subprocess.PIPE,shell=1)
		stdout, stderr = process.communicate()
		if stderr: print(stderr)
		
		stdout = stdout.decode("cp850")
		for x in stdout.splitlines():
			yield x.strip()
	except:
		pass
	
def Pause():
	os.system("pause")

def Clearterminal():
	os.system("cls")

class to_string:
	__BYTE_HELPER = [[math.pow(1024,i),name] for i,name in enumerate(["bytes","KB","MB","GB","TB"])]
	__BYTE_HELPER.reverse()
	@classmethod
	def bytes(cls,number_of_bytes):
		number_of_bytes = float(number_of_bytes)
		if not number_of_bytes:
			return "0 bytes"
		for a in cls.__BYTE_HELPER:
			n = number_of_bytes/a[0]
			if n >= 1:
				return "{a} {b}".format(a=round(n,2),b=a[1])
	@classmethod
	def sec(cls,s):
		h = math.floor(s/3600)
		s -= h*3600
		m =  math.floor(s/60)
		s -= m*60
		return "%02d:%02d:%02d" % (h,m,s)

	@classmethod
	def unix(cls,i):
		return datetime.utcfromtimestamp(i).strftime('%Y-%m-%d %H:%M:%S')


def to_int(n):
	try:
		retval = int(n)
		return retval
	except:
		return None

def good_list_index(index,len_ar):
	return max(0, min(index, len_ar - 1))

def all_true(funcs:List[callable],*a,**b):
	for f in funcs:
		if not f(*a,**b):
			return False
	return True

def first_item(condition,ar:List[any]):
	for i in range(len(ar)):
		if condition(ar[i],i):
			return ar[i]

def first_index(condition,ar:List[any]):
	for i in range(len(ar)):
		if condition(ar[i],i):
			return i
	return None

def cb_func(*a,**b):
	print(f"a : {a}")
	print(f"b : {b}")

def convert_func(func,no_error:bool=False,condition=None,no_para:bool=False):
	t = {k : v for k,v in locals().items() if k != "func"}
	
	if t.get('no_error',False):
		t['no_error'] = False
		def ret_val(*a,**b):
			try:
				ret = func(*a,**b)
				return ret
			except:
				pass
		return convert_func(ret_val,**t)
	if t.get('condition',None):
		condition_func = t.get("condition")
		
		def ret_val(*a,**b):
			if condition_func(*a,**b):
				ret = func(*a,**b)
				return ret
		t['condition'] = None
		return convert_func(ret_val,**t)

	if t.get('no_para',None):
		t['no_para'] = False
		def ret_val(*a,**b):
		
			ret = func()
			return ret

		return convert_func(ret_val,**t)
	


	return func

def convert_func_no_error(f):
	def ret_func(*a,**b):
		try:
			retval = f(*a,**b)
			return retval
		except:
			pass

	return ret_func

def convert_func_condition(f,condition):
	def ret_func(*a,**b):
		if condition(*a,**b):
			return f(*a,**b)
	
	return ret_func

def convert_func_strip_para(f):
	def ret_func(*a,**b):
		return f()
	
	return ret_func

def convert_func_args_kwargs(f,args,kwargs):
	def ret_func():
		return f(*args,**kwargs)
	
	return ret_func


def ar_flatten(ar):
	new_ar = []
	for x in ar:
		new_ar += x

	return new_ar



def convert_func_pff(func,no_error:bool=False,condition=None,no_para:bool=False,kwargs=dict(),args=[]):
	ret_func = func
	if no_error:
		ret_func = convert_func_no_error(ret_func)

	if condition:
		ret_func = convert_func_condition(ret_func,condition)

	if no_para:
		ret_func = convert_func_strip_para(ret_func)


	ret_func = convert_func_args_kwargs(ret_func,args,kwargs)

	return ret_func
	
import sys


def export(fn):
	mod = sys.modules[fn.__module__]
	if hasattr(mod, '__all__'):
		mod.__all__.append(fn.__name__)
	else:
		mod.__all__ = [fn.__name__]
	return fn

#									blabla packed

class Ob:
	properties_set_all  = properties_set_all
	properties_keys_ar  = properties_keys_ar
	properties_keys_tab = properties_keys_tab



class String:
	same_len_ar = same_len_ar
	random  	= random
	between 	= between
	fill_to_len = fill_to_len

class Index:
	first_index 	= first_index
	first_item  	= first_item
	all_true 		= all_true
	good_list_index = good_list_index
	

	@staticmethod
	def first_index_without_index(condition,ar:List[any]):
		for i in range(len(ar)):
			if condition(ar[i]):
				return i
		return None
	@staticmethod
	def first_item_without_index(condition,ar:List[any]):
		for i in range(len(ar)):
			if condition(ar[i]):
				return ar[i]



import datetime, http, math, os, random, traceback
from operator 			import itemgetter
from queue 				import Queue
from threading 			import Thread
from time 				import sleep, time
from typing 			import List
from urllib.request 	import Request, urlopen
from ..util 			import to_int, to_string
from ..wrapper 			import jvWrapper


class DownloadFileInternal:
	block_size            : int          = 0
	nthreads              : int          = 0
	url                   : str          = None
	fp                    : str          = None
	size                  : int          = 0
	accept_ranges         : bool         = False

	info_thread           : Thread     
	info_write_speed = 0
	info_download_speed = 0

	writer_bytes_written  : int          = 0
	writer_blocks_written : int          = 0
	writer_q              : Queue        = None
	writer_thread         : Thread      

	dblock_bytes_loaded   : int          = 0
	dblock_success        : int          = 0
	dblock_total          : int          = 0
	dblock_failed         : int          = 0
	dblock_q              : Queue        = None
	dblock_threads        : List[Thread] = []

	status_running        : bool         = False

	def __init__(self,url:str,fp:str,block_size_kb:int=500,nthreads:int=5):
		self.url             = url
		self.fp              = fp
		self.block_size      = block_size_kb * 1024
		self.nthreads        = nthreads
		
		self.info_init()
		self.dlbock_init()
		self.status_init()
		self.writer_init()

	def status_init(self):
		self.status_running  = False


	def dlbock_init(self):
		self.dblock_q        = Queue()
		self.dblock_threads  = [Thread(target=self.dblock_thread_func,daemon=True) for i in range(self.nthreads)]


	def dblock_reset(self):
		self.dblock_bytes_loaded = 0
		self.dblock_success      = 0
		self.dblock_total        = 0
		self.dblock_failed       = 0

	def dblock_start(self):
		self.dblock_reset()
		self.dblock_total = self.dblock_q.qsize()
		for t in self.dblock_threads:
			t.start()

	def dblock_thread_func(self):
		def download_block(index=0,failed=0):
			def action_success():
				self.dblock_success += 1
			def action_failed():
				if failed < 2:
					self.dblock_q.put(dict(index=index,failed=failed+1))
					return
				else:
					self.dblock_failed += 1
			
			block_start, block_stop = ((index*self.block_size), min( (self.size - 1) , ((index + 1) * self.block_size - 1) ) )

			data_len = block_stop - block_start + 1
			request    = None

			for attemp in range(5):
				try:
					request                  = Request(self.url)
					request.headers["Range"] = 'bytes=%d-%d'%(block_start,block_stop)
					break
				except:
					print(traceback.format_exc())
					sleep(3)
			
			if request is None:
				return action_failed()
			for attemp in range(10):
				try:
					data = urlopen(request)
					data = data.read()
					self.dblock_bytes_loaded += data_len
					self.writer_q.put(dict(index=index,data=data))
					return action_success()
				except (http.client.IncompleteRead) as e:
					data_len_to_do = (block_stop - block_start + 1)
					self.dblock_bytes_loaded += data_len_to_do
					data_to_write = e.partial + str.encode('0'*( data_len_to_do - len(e.partial)))
					self.writer_q.put(dict(index=index,data=data_to_write))
					
		
					return action_success()
				except:
					print(traceback.format_exc())
					sleep(2)
		
			return action_failed()
		
		while self.status_running and ( not self.status_finished ):
			while self.dblock_q.empty():
				if self.dblock_finished:
					return
				sleep(2)
			
			todo        = self.dblock_q.get()
			download_block(**todo)
	
	def writer_init(self):
		self.writer_q        = Queue()
		self.writer_thread   = Thread(target=self.writer_thread_func,daemon=True)

	def writer_reset(self):
		self.writer_bytes_written  = 0
		self.writer_blocks_written = 0

	def writer_start(self):
		self.writer_reset() #!
		self.writer_thread.start()

	def writer_thread_func(self):
		def try_writing():
			todo = []
			while not self.writer_q.empty():
				q = self.writer_q.get()
				todo.append(q)
			
			if not todo:
				return
			todo.sort(key=lambda x: x.get('index'))

			with open(self.fp, "r+b") as f:
				cursor_now  = 0
				for td in todo:
					index,data 		   = itemgetter('index','data')(td)
					cursor_destination = index * self.block_size
					if (cursor_delta := cursor_destination-cursor_now) > 0:
						f.seek(cursor_delta,1)
					f.write(data)
					writtenNow    = len(data)
					cursor_now    = cursor_destination+writtenNow
					self.writer_bytes_written += writtenNow
					self.writer_blocks_written += 1

		
		while self.status_running and ( not self.status_finished ):
			try_writing()
			sleep(1)
			
	def info_init(self):
		self.info_thread     = Thread(target=self.info_thread_func,daemon=True)

	def info_start(self):
		self.info_thread.start()



	def info_thread_func(self):

		self.info_write_speed          = 0
		self.info_download_speed       = 0
		sleep_time = 1

		def get_data_now():
			return dict(
				t=time(),
				written=self.writer_bytes_written,
				downloaded=self.dblock_bytes_loaded
			)
		data = []
		data.append(get_data_now())
		sleep(1)
		while self.status_running and ( not self.status_finished ):
			data.append(get_data_now())
			data = data[-5:]

			first_data_entry = data[0]
			last_data_entry = data[-1]
			
			delta = {k : last_data_entry[k] - first_data_entry[k] for k in ['t','written','downloaded']}
			
			self.info_write_speed		= delta.get('written') / delta.get('t')
			self.info_download_speed	= delta.get('downloaded') / delta.get('t')

			sleep_time_more = max(0, sleep_time - (time() - last_data_entry.get('t')))
			if sleep_time_more:
				sleep(sleep_time_more)

	

	
	def reload(self):
		def reload_header():
			self.size          = 0
			self.accept_ranges = False
			for attempt in range(5):
				try:
					open_response = urlopen(self.url,timeout=10)
					self.size = to_int(open_response.headers.get('Content-Length',0))
					if not self.size:
						return False
					self.accept_ranges  = open_response.headers.get('accept-ranges','no') == 'bytes'
					return True
				except:
					sleep(1)
			
			return False
	
		def reload_local_file():
			def write_empty_file_if_not_exist():
				file_already_started 			   = os.path.isfile(self.fp) and (os.path.getsize(self.fp) == self.size)
				if file_already_started:
					return
				
				bytes_written 					   = 0
				bytes_to_write                     = self.size
				empty_str_20mb_len                 = 1024*1024*20
				empty_str_20mb                     = str.encode('0'*empty_str_20mb_len)
				with open(self.fp,"wb") as f:
					while bytes_written + empty_str_20mb_len <= bytes_to_write:
						f.write(empty_str_20mb)
						bytes_written += empty_str_20mb_len
					
					bytes_to_write_more = bytes_to_write -  bytes_written
					if bytes_to_write_more:

						f.write(str.encode('0'*bytes_to_write_more))
						

				sleep(1)

			def get_block_indexes_to_download():
				block_size             = self.block_size
				empty_block_str        = str.encode('0'*block_size)
				block_index            = 0
				indexes_to_download    = []
				blocks_already_written = 0
				
				with open(self.fp,"rb") as f:
					data = f.read(block_size)
					while data:
						data_len  = len(data)
						do_append = ((data_len == block_size) and (data == empty_block_str)) or ((data_len != block_size) and (data == str.encode('0'*data_len)))
						if do_append:
							indexes_to_download.append(block_index)
						else:
							blocks_already_written 	+= 1

						block_index += 1
						data 		= f.read(block_size)


				return indexes_to_download	

			def set_block_indexes_to_download(indexes):
				for i in indexes:
					self.dblock_q.put((dict(index=i,failed=0)))

			write_empty_file_if_not_exist()
			block_indexes_to_download = get_block_indexes_to_download()


			set_block_indexes_to_download(block_indexes_to_download)
		reload_header()
		reload_local_file()

	def stop(self):
		self.status_running = False
		return True
	
	def start(self):
		self.reload()
		self.status_running = True
		funcs_to_do = [self.dblock_start,self.writer_start,self.info_start]
		for t in funcs_to_do:
			t()
		

	def reset(self):
		print(f'Reset not yet implemented')

	def move(self,new_fp):
		print(f'Move not yet implemented!')
	
	@property
	def writer_finished(self) ->bool:
		return self.writer_blocks_written >= (self.dblock_total - self.dblock_failed)

	@property
	def dblock_finished(self) ->bool:
		return self.dblock_total == (self.dblock_success + self.dblock_failed)

	
	@property
	def status_finished(self) -> bool:
		return self.writer_finished and self.dblock_finished
	


class _DownloadHelperData:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	@property
	def total(self):						return self.__downloader.size
	
	@property
	def downloaded(self):					return self.__downloader.dblock_bytes_loaded

	@property
	def written(self):						return self.__downloader.writer_bytes_written

	@property
	def to_download(self):					return self.total - self.downloaded

	@property
	def to_write(self):						return self.total - self.written

	@property
	def speed(self):						return self.__downloader.info_download_speed



class _DownloadHelperDataHuman:
	__data : _DownloadHelperData
	def __init__(self,data):
		self.__data = data

	@property
	def total(self) -> str:				return to_string.bytes(self.__data.total)
	
	@property
	def downloaded(self) -> str:		return to_string.bytes(self.__data.downloaded)
	
	@property
	def written(self) -> str:			return to_string.bytes(self.__data.written)
	
	@property
	def to_download(self) -> str:		return to_string.bytes(self.__data.to_download)
	
	@property
	def to_write(self) -> str:			return to_string.bytes(self.__data.to_write)
	
	@property
	def speed(self) -> str:				return to_string.bytes(self.__data.speed) + '/s'



class _DownloadHelperTime:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader
	
	@property
	def remaining(self):
		speed = self.__downloader.info.speed
		to_download = self.__downloader.src.size - self.__downloader.blocks.bytes_downloaded
		if speed == 0:
			return -1
		return to_download / speed
	
	@property
	def elapsed(self):				return time() - self.__downloader.time_started

	@property
	def started(self):				return self.__downloader.time_started

	@property
	def ended(self):				return self.__downloader.time_ended



class _DownloadHelperTimeHuman:
	__time : _DownloadHelperTime
	def __init__(self,timeOb):
		self.__time = timeOb

	@property
	def remaining(self) -> str:		return to_string.sec(self.__time.remaining)
	
	@property
	def elapsed(self) -> str:		return to_string.sec(self.__time.elapsed)
	
	@property
	def started(self) -> str:		return to_string.unix(self.__time.started)
	
	@property
	def ended(self) -> str:			return to_string.unix(self.__time.ended)

class _DownloadHelperError:
	src          : str                 
	download     : str                 
	write        : str                 
	__downloader : DownloadFileInternal
	
	def __init__(self,downloader):
		self.__downloader = downloader
	@property
	def has(self):
		return self.src or self.download or self.write
	
class _DownloadHelperStatus:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	@property
	def finished(self) -> bool:				return self.__downloader.status_finished
	
	@property
	def running(self) -> bool:				return self.__downloader.status_running

	@property
	def stopped(self) -> bool:				return not self.__downloader.status_running

	@property
	def started(self) -> bool:				return self.__downloader.status_finished or self.__downloader.status_running

	@property
	def error(self) -> bool:
		return False
#		return self.__downloader.has_errors
	

class _DownloadHelperControl:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	def start(self):							return self.__downloader.start()
	def stop(self):								return self.__downloader.stop()
	def reset(self):							return self.__downloader.reset()
	def move(self,new_fp=r'C:\temp\temp.tmp'):	return self.__downloader.move(new_fp)



class Download:
	__downloader : DownloadFileInternal              
	data         : _DownloadHelperData     
	dataHuman    : _DownloadHelperDataHuman
	time         : _DownloadHelperTime     
	timeHuman    : _DownloadHelperTimeHuman
	error        : _DownloadHelperError    
	status       : _DownloadHelperStatus   
	control      : _DownloadHelperControl  

	def __init__(self,url:str,fp:str, block_size_kb : int = 400, max_threads : int = 10):
		self.__downloader = DownloadFileInternal(url,fp,block_size_kb,max_threads)
		self.data         = _DownloadHelperData(self.__downloader)
		self.dataHuman    = _DownloadHelperDataHuman(self.data)
		self.time         = _DownloadHelperTime(self.__downloader)
		self.timeHuman    = _DownloadHelperTimeHuman(self.time)
		self.error        = _DownloadHelperError(self.__downloader)
		self.status       = _DownloadHelperStatus(self.__downloader)
		self.control 	  = _DownloadHelperControl(self.__downloader)



from ..util 		import download,string
from ..fs 			import JDir,JFile
from ..wrapper 		import jvWrapper,jvWrapperSuper
from typing 		import List
from ..jv 			import jv



DEBUG = False


class CONSTANT:
	packagename     : str  
	version         : str  
	fConfigFile     : JFile
	fSetupFile      : JFile
	dPythonPackages : JDir 
	dPythonPackage  : JDir 
	dPackage        : JDir 
	dSettings       : JDir 
	dMainDir        : JDir 
	dDist			: JDir

	fBatBuild		: JFile
	fBatUpload	    : JFile



@jvWrapper()
class CONSTANT_SETTER(jvWrapperSuper):
	current_function_name : str
	def __init__(self):
		pass

	def init(self):
		def add():
			def decorater(f):
				self.current_function_name = f.__name__
				if not f():
					self.error('miauw')
			return decorater
		
		def must_exist(d):
			return d and d.exists
		
		def try_setting(*conditions,key=None,value=None):
			pass
		
		
		@add()
		def dMainDir():
			CONSTANT.dMainDir = JFile(__file__).parent_dir_ob
			return CONSTANT.dMainDir.exists
		@add()
		def dSettings():
			CONSTANT.dSettings = CONSTANT.dMainDir.copyob.down('settings')
			return CONSTANT.dSettings.exists
		
		@add()
		def fConfigFile():
			CONSTANT.fConfigFile = CONSTANT.dSettings.file("config.json")
			return CONSTANT.fConfigFile.exists
		
		@add()
		def fSetupFile():
			CONSTANT.fSetupFile = CONSTANT.dSettings.file("setup.py")
			return CONSTANT.fSetupFile.exists
		
		@add()
		def version():
			CONSTANT.version = CONSTANT.fConfigFile.reader.json().get('version')
			return True
		
		@add()
		def packagename():
			CONSTANT.packagename = CONSTANT.fConfigFile.reader.json().get('packagename')
			return True
		
		@add()
		def dPythonPackage():
			CONSTANT.dPythonPackage = JFile(jv.__file__).parent_dir_ob
			return CONSTANT.dPythonPackage.exists
		
		@add()
		def dPythonPackages():
			CONSTANT.dPythonPackages = CONSTANT.dPythonPackage.parent_dir_ob
			return CONSTANT.dPythonPackages.exists
		
		@add()
		def dPackage():
			CONSTANT.dPackage = CONSTANT.dMainDir.copyob.down('data','current')
			return True
		
		@add()
		def dDist():
			CONSTANT.dDist = CONSTANT.dPackage.copyob.down('dist')
			return True	
		
		@add()
		def bats():
			dBat = CONSTANT.dSettings.copyob.down('bats')
			CONSTANT.fBatBuild = dBat.file('build.bat')
			CONSTANT.fBatUpload = dBat.file('upload.bat')
			return CONSTANT.fBatBuild.exists and CONSTANT.fBatUpload.exists
	
	def error(self,*a):
		self.jvlog.critical(self.current_function_name,*a)
		exit()
	
	def set_next_version_from_site(self):
		url = f"https://pypi.org/project/{CONSTANT.packagename}/"

		def get_html():
			if DEBUG:
				return CONSTANT.dSettings.file("test.html").reader.text(encoding="utf8")
			return download.html(url)

		html = get_html()
		if not html:
			return self.error("set_next_version_from_site", "failed to download",url)
			
		version = string.between(html,'package-header__name">','</h1>')
		if not version:
			return self.error("set_next_version_from_site", "failed to read between",html)

		version = version.replace(CONSTANT.packagename,'').strip()

		old_version = CONSTANT.version
		CONSTANT.version = "0.0.{i}".format(i=int(version.split('.')[-1]) + 1)


		keys = list(CONSTANT.fConfigFile.reader.json().keys())
		self.jvlog.info(f"Version {old_version} -> {CONSTANT.version}")

		CONSTANT.fConfigFile.writer.json({k : getattr(CONSTANT,k) for k in keys})
		

	@classmethod
	def do(cls):
		c = cls()
		c.init()
		c.set_next_version_from_site()

CONSTANT_SETTER.do()



@jvWrapper()
class PackageUploader(jvWrapperSuper):

	current_function_name : str
	def __init__(self):
		pass
	
	def error(self,*a):
		self.jvlog.critical(self.current_function_name,*a)
		exit()


	def init(self):
		def add():
			def decorater(f):
				self.current_function_name = f.__name__
				if not f():
					self.error('miauw')
			return decorater
		

		@add()
		def del_dPackage():
			return CONSTANT.dPackage.delete_full()
		@add()
		def create_dPackage():
			return CONSTANT.dPackage.create()
		@add()
		def create_dPackageDirs():
			for d in list(filter(lambda x : x.lname != "__pycache__",CONSTANT.dPythonPackage.reader.dirs_ob(deep=True))):
				d_to = CONSTANT.dPackage.copyob.down(d.rel(CONSTANT.dPythonPackages))
				if not d_to.create():
					return False
			return True
		@add()
		def copy_dPackageFiles():
			def file_filter(f:JFile):
				if f.fp.lower().find('__pycache__') != -1:
					return False
				if f.lex == 'cfg':
					return False
				return True
			
			for f_from in list(filter(file_filter,CONSTANT.dPythonPackage.reader.files_ob(deep=True))):
				f_parent_dir = f_from.parent_dir_ob
				d_to_parent_dir = CONSTANT.dPackage.copyob.down(f_parent_dir.rel(CONSTANT.dPythonPackages))
				f_to = f_from.copyob
				f_to.parent_dir_fp = d_to_parent_dir.fp
				if not f_from.copy(f_to.fp,True):
					return False
			return True
		@add()
		def copy_build_files():
			to_file = CONSTANT.fSetupFile.copyob
			to_file.parent_dir_fp = CONSTANT.dPackage.fp
			return CONSTANT.fSetupFile.copy(to_file.fp,True)


		@add()
		def del_dist():
			return CONSTANT.dDist.delete_full()
		
		bat_start_options = dict(wait=True,same_window=True,start_dir=CONSTANT.dPackage.fp)
		bat_start_options = dict(wait=True,same_window=False,start_dir=CONSTANT.dPackage.fp)


		@add()
		def create_build():

			CONSTANT.fBatBuild.start(title=f"{CONSTANT.packagename} builder",**bat_start_options)
			return True
		
		@add()
		def upload():
			CONSTANT.fBatUpload.start(title=f"{CONSTANT.packagename} uploader",**bat_start_options)
			return True



pu = PackageUploader()

pu.init()



from ..util.ob import copy_tab
from typing import List
def string_replacer(s,tab,**t):
	for k,v in tab.items():
		if v is None:
			continue
		s = s.replace(f'[{k}]',v)
	return s

class IHasReplacement_tab:
	@property
	def replacement_tab(self):
		return dict()





from __future__ 			import annotations
import 					   curses, fnmatch, math
from time 			import time,sleep
from ..menu 		import Menu,MenuItem,TerminalPart,Ctrl,LineConstructor,Event,LinesArea,LineArea
from ...util 		import Thread

class Addon:
	def __init__(self,ctrl:Ctrl):
		self.ctrl = ctrl
	
		self.register_listeners()
	
	@property
	def addstr(self) -> curses.window.addstr:
		return self.cur.win.addstr
	
	@property
	def cur(self):
		return self.ctrl.cur
	
	@property
	def on_key(self):
		return self.ctrl.on_key
	
	@property
	def event(self):
		return self.ctrl.event
	
	@property
	def activeMenu(self):
		return self.ctrl.active_menu
	
	@property
	def selectedItem(self):
		return self.ctrl.selected_item
	
	def register_listeners(self):
		pass


class FirstLetter(Addon):
	total_received : str 
	last_done      : int 
	enabled        : bool

	def reset(self):
		self.total_received = ""
		self.last_done      = 0
	def register_listeners(self):
		self.total_received = ""
		self.last_done      = 0
		self.enabled        = True
		event               = self.event
		on_key              = self.on_key
		
		@event.focused_on_list_changed()
		def focused_on_list_changed(ev:Event.FocusedOnListChanged):
			self.reset()
			self.enabled = ev.focused_now

		@event.selected_changed_manual()
		def changed(ev:Event.Event):
			self.reset()
		
		@event.char_received()
		def on_char_received(ev:Event.CharReceived):
			if not self.enabled:
				return
			
			if (time() - self.last_done) > 1:
				self.total_received = ""
			received = ev.received
			if received.name is None and received.c:
				self.total_received += received.c
			else:
				self.total_received = ''
			self.last_done = time()
			menu = self.activeMenu
			if not menu:
				return
			
			if self.total_received:
				for item in menu.items_sorted_by_text(menu.f_items):
					if item.text.lower().startswith(self.total_received):
						return menu.set_filter_selected_not_manual(item.filter_index)

class Mousedownwatcher:
	def __init__(self,scrollbar:Scrollbar):
		self.scrollbar = scrollbar
		self.register_listeners()
		self.last_event = 0
	@property
	def ctrl(self):
		return self.scrollbar.ctrl

	@property
	def addstr(self) -> curses.window.addstr:
		return self.cur.win.addstr
	
	@property
	def cur(self):
		return self.ctrl.cur
	
	@property
	def on_key(self):
		return self.ctrl.on_key
	
	@property
	def event(self):
		return self.ctrl.event
	
	@property
	def activeMenu(self):
		return self.ctrl.active_menu
	
	@property
	def selectedItem(self):
		return self.ctrl.selected_item

	@property
	def area_plus(self):
		return self.scrollbar.area_plus
	
	@property
	def area_min(self):
		return self.scrollbar.area_min

	def mouse_down_thread(self,func,last_event):
		print(f"mouse_down_thread started {func.__name__}")
		func()
		sleep(0.5)
		while last_event == self.last_event:
			func()
			sleep(0.5)
		print(f"mouse_down_thread stop {func.__name__}")

	def register_listeners(self):
		events = self.ctrl.event
		cur    = self.cur

		@events.mouse_down()
		def mc(ev:Event.MouseClick):
			print(f"received mouse_down")
			last_event      = time()
			self.last_event = last_event
			pos             = ev.coord
			m = self.activeMenu
			if not m:
				return
			
			if self.area_plus.pos_is_in(pos):
				Thread(self.mouse_down_thread,(m.page_down,last_event))
			if self.area_min.pos_is_in(pos):
				Thread(self.mouse_down_thread,(m.page_up,last_event))

		@events.mouse_up()
		def mc(ev:Event.MouseClick):
			print(f"received mouse up")
			self.last_event = time()
	

class Scrollbar(Addon):
	area_plus        : LinesArea       
	area_min         : LinesArea       
	mousedownwatcher : Mousedownwatcher
	def print_scrollbar(self,min_y,max_y,x):
		color                 = self.cur.color.selected
		addstr                = self.cur.win.addstr
		self.mousedownwatcher = Mousedownwatcher(self)
		def print_button(start_y,s,area:LinesArea):
			texts = ["   ","   "]
			for i,line in enumerate(area.lines):
				line         : LineArea = line
				line.y                  = start_y+i
				line.start_x            = x-3
				line.stop_x             = x
				addstr(line.y,line.start_x,texts[i],self.cur.color.blue)
		
		print_button(min_y,"-",self.area_min)
		print_button(max_y - 1,"+",self.area_plus)

		def print_scrollbar(start_y,stop_y,start_x):
			m = self.activeMenu
			length 		 = max_y - min_y - 3
			position_now = math.ceil( ((m.f_selected_index + 1)/ len(m.f_items)) * length + min_y + 1)
			for line in range(start_y,stop_y):
				col = self.cur.color.mag
				if position_now == line:
					col = self.cur.color.blue
				addstr(line,start_x," ",col)
		
		print_scrollbar(min_y + 2, max_y - 1, x - 2)
	
	def register_listeners(self):
		def get_dummy_lines_area():
			x = LinesArea()
			for i in range(2):
				x.append(LineArea(0,0,0))
			return x
		self.area_plus = get_dummy_lines_area()
		self.area_min  = get_dummy_lines_area()
		events         = self.ctrl.event
		cur            = self.cur

		@events.mouse_clicked()
		def mc(ev:Event.MouseClick):
			pos = ev.coord

			m = self.activeMenu
			if not m:
				return
			
			if self.area_plus.pos_is_in(pos):
				m.page_down()
			if self.area_min.pos_is_in(pos):
				m.page_up()

		@events.after_print()
		def build_print(ev:Event.Event):
			old_cursor = cur.cursor_position
			min_y,max_y = self.ctrl.terminal_items.line_range
			if min_y is None:
				return
			max_x = cur.max_x
			self.print_scrollbar(min_y,max_y,max_x)
			cur.win.move(old_cursor.y,old_cursor.x)

class Searchbox(Addon):
	area_input  					: LinesArea
	searching                       : bool = False
	first_build_since_input_changed : bool = False
	__search_input                  : str  = ""
	width : int = 20
	height : int = 1

	@property
	def search_input(self):
		return self.__search_input

	@search_input.setter
	def search_input(self,value):
		if value != self.__search_input:
			self.__search_input = value
			if value:
				self.first_build_since_input_changed = True
	

	def on_activate(self):
		self.searching = True
		cur       = self.cur
		last_line = self.area_input.lines[-1]

		cur.win.move(last_line.y,last_line.start_x)
		curses.curs_set(2)
		cur.win.refresh()

	
	def stop_searching(self):
		self.searching = False
	
	def on_searching(self):
		cur       = self.cur
		last_line = self.area_input.lines[-1]

		self.addstr(last_line.y,last_line.start_x,self.search_input,cur.color.selected)
		
		curses.curs_set(2)
		cur.win.refresh()

	
	def print_search_input(self,window_width):
		if len(self.area_input.lines) != self.height:
			self.area_input.lines = self.area_input.lines[:self.height]
			self.area_input.add_dummy(self.height - len(self.area_input.lines))
		
				
		
		addstr = self.addstr
		for i in range(self.height):
			addstr(i,window_width - self.width,"_"*self.width,self.cur.color.fullcyan)
			self.area_input.lines[i].y = i
			self.area_input.lines[i].start_x = window_width - self.width
			self.area_input.lines[i].stop_x = window_width

	def register_listeners(self):
		self.area_input                      = LinesArea()
		self.__search_input                  = ""
		self.searching                       = False
		self.first_build_since_input_changed = False
		events                               = self.ctrl.event
		on_key                               = self.ctrl.on_key
		cur                                  = self.cur

		for i in range(self.height):
			self.area_input.append(LineArea(0,0,0))
			self.area_input.append(LineArea(0,0,0))


		def leave_search():

			self.searching = False

		@self.ctrl.filter
		def filter(mi:MenuItem):
			if not self.searching:
				return True
			if not self.search_input:
				return True
			return fnmatch.fnmatch(mi.text,'*' + self.search_input + '*')

		@on_key(curses.KEY_F2)
		def toggle_searching(ctrl:Ctrl):
			self.search_input 		= ''
			self.searching 			= not self.searching
			ctrl.focused_on_list 	= not self.searching
		
		@on_key(curses.KEY_BACKSPACE)
		def back_space(ctrl:Ctrl):
			if self.searching:
				self.search_input = self.search_input[:-1]
				if not self.search_input:
					self.stop_searching()
		
		@events.char_received()
		def ok(ev:Event.CharReceived):
			inp = ev.received
			if self.searching and (inp.c) and (not inp.name):
				self.search_input = self.search_input + inp.c
			
		@events.after_print()
		def build_print(ev:Event.Event):
			old_cursor = cur.cursor_position
			self.print_search_input(cur.max_x)
			cur.win.move(old_cursor.y,old_cursor.x)
			if self.searching:
				self.on_searching()

		@events.mouse_clicked()
		def searchbox_mc(ev:Event.MouseClick):
			pos = ev.coord
			m = self.activeMenu
			if not m:
				return
			
			if self.area_input.pos_is_in(pos):
				self.on_activate()
		
		@events.building_print()
		def build_print(ev:Event.Event):
			first_build = self.first_build_since_input_changed
			self.first_build_since_input_changed = False
			if not self.searching:
				return
			
			menu = self.activeMenu
			if not menu:
				return
			
			bottom = self.ctrl.terminal_bot
			bottom.create_block(LineConstructor(empty=2))
			@bottom.wrapper(LineConstructor(text=f'#results : {len(menu.f_items)}/{len(menu.items)}'),LineConstructor(text=f'Searching : {self.search_input}',add_line_break=False))
			def kk(*a,**b):
				print(f"a {a}")
				print(f"b {b}")
			
			if first_build:
				menu.d_first_f_index = 0

class Searcher(Addon):
	searching : bool = False

	first_build_since_input_changed : bool = False
	__search_input : str = ""

	@property
	def search_input(self):
		return self.__search_input

	@search_input.setter
	def search_input(self,value):
		if value != self.__search_input:
			self.__search_input = value
			if value:
				self.first_build_since_input_changed = True
	
	def register_listeners(self):
		self.__search_input = ""
		self.searching = False
		self.first_build_since_input_changed = False

		events = self.ctrl.event
		on_key = self.ctrl.on_key
		
		@self.ctrl.filter
		def filter(mi:MenuItem):
			if not self.searching:
				return True
			if not self.search_input:
				return True
			return fnmatch.fnmatch(mi.text,'*' + self.search_input + '*')
		
		@on_key(curses.KEY_F2)
		def toggle_searching(ctrl:Ctrl):
			self.search_input 		= ''
			self.searching 			= not self.searching
			ctrl.focused_on_list 	= not self.searching
		
		@on_key(curses.KEY_BACKSPACE)
		def back_space(ctrl:Ctrl):
			if self.searching:
				self.search_input = self.search_input[:-1]
		
		@events.char_received()
		def ok(ev:Event.CharReceived):
			inp = ev.received
			if self.searching and (inp.c) and (not inp.name):
				self.search_input = self.search_input + inp.c
			
		@events.building_print()
		def build_print(ev:Event.Event):
			first_build = self.first_build_since_input_changed
			self.first_build_since_input_changed = False
			if not self.searching:
				return
			
			menu = self.activeMenu
			if not menu:
				return
			
			bottom = self.ctrl.terminal_bot

			bottom.create_block(LineConstructor(empty=2))
			@bottom.wrapper(LineConstructor(text=f'#results : {len(menu.f_items)}/{len(menu.items)}'),LineConstructor(text=f'Searching : {self.search_input}',add_line_break=False))
			def kk(*a,**b):
				print(f"a {a}")
				print(f"b {b}")
			
			if first_build:
				menu.d_first_f_index = 0

class Searcher(Addon):
	searching : bool = False

	first_build_since_input_changed : bool = False
	__search_input : str = ""

	@property
	def search_input(self):
		return self.__search_input

	@search_input.setter
	def search_input(self,value):
		if value != self.__search_input:
			self.__search_input = value
			if value:
				self.first_build_since_input_changed = True
	
	def register_listeners(self):
		self.__search_input = ""
		self.searching = False
		self.first_build_since_input_changed = False

		events = self.ctrl.event
		on_key = self.ctrl.on_key
		
		@self.ctrl.filter
		def filter(mi:MenuItem):
			if not self.searching:
				return True
			if not self.search_input:
				return True
			return fnmatch.fnmatch(mi.text,'*' + self.search_input + '*')
		
		@on_key(curses.KEY_F2)
		def toggle_searching(ctrl:Ctrl):
			self.search_input 		= ''
			self.searching 			= not self.searching
			ctrl.focused_on_list 	= not self.searching
		
		@on_key(curses.KEY_BACKSPACE)
		def back_space(ctrl:Ctrl):
			if self.searching:
				self.search_input = self.search_input[:-1]
		
		@events.char_received()
		def ok(ev:Event.CharReceived):
			inp = ev.received
			if self.searching and (inp.c) and (not inp.name):
				self.search_input = self.search_input + inp.c
			
		@events.building_print()
		def build_print(ev:Event.Event):
			first_build = self.first_build_since_input_changed
			self.first_build_since_input_changed = False
			if not self.searching:
				return
			
			menu = self.activeMenu
			if not menu:
				return
			
			bottom = self.ctrl.terminal_bot

			bottom.create_block(LineConstructor(empty=2))
			@bottom.wrapper(LineConstructor(text=f'#results : {len(menu.f_items)}/{len(menu.items)}'),LineConstructor(text=f'Searching : {self.search_input}',add_line_break=False))
			def kk(*a,**b):
				print(f"a {a}")
				print(f"b {b}")
			
			if first_build:
				menu.d_first_f_index = 0


from jivi.event import EventWithKeyWithParent
from typing import List
import random,traceback,math
from ..constant import ALL_SHAPES,CLine,ScoreSetting,Data
from itertools import product
from time import time

def get_random_shape_index():
	return random.randint(0,len(ALL_SHAPES) - 1)

def ar_passes_all(condition,ar=None,as_function=False,inverse=False):
	if as_function:
		def ret_val(ar):
			return ar_passes_all(condition,ar=ar,inverse=inverse)
		return ret_val
	if inverse:
		return not ar_passes_all(condition,ar=ar)
	for x in ar:
		if not condition(x):
			return False
	return True

class Block:
	def __init__(self,x:int,y:int,shape_index:int=None,rotation:int=0):
		self.x           = x
		self.y           = y
		self.shape_index = get_random_shape_index if shape_index is None else shape_index
		self.set_rotation(rotation)
	def set_rotation(self,rotation):	self.__rotation = rotation % 4
	def get_positions(self):			return [[x+self.x,y+self.y] for x,y in ALL_SHAPES[self.shape_index][self.__rotation]]
	def move(self,x=0,y=0,rotate=0): 	return self.__class__(shape_index=self.shape_index,x=self.x + x,y=self.y + y,rotation=self.__rotation + rotate)
	def __repr__(self):					return f'{self.__class__.__name__} ({self.x},{self.y})'



class CtrlElement:
	last_change : int = 0
	def __init__(self):
		self.changed()
	
	def changed(self):
		self.last_change = time()


	def get_data(self):
		pass

class Score(CtrlElement):
	def __init__(self):
		self.score       = 0
		self.lines       = 0
		self.level       = 0
		super().__init__()
	def _update_level(self):
		while self.score > ScoreSetting.score_at_level[self.level]:
			self.level += 1
	
	def add_lines(self,to_add:int):
	
		self.lines += to_add
		self.score += ScoreSetting.score_for_lines_made[to_add]
		self._update_level()
		self.last_change = time()

	def get_data(self) -> Data.Score:
		return Data.Score(level=self.level,score=self.score,lines=self.lines)


class BlockContainer(CtrlElement):
	_shape : int = -1
	@property
	def shape(self):
		return self._shape
	
	@shape.setter
	def shape(self,n):
		self._shape = n
		self.changed()
	
	def get_data(self) -> Data.BlockContainer:
		if self._shape == -1:
			return Data.BlockContainer()
		return Data.BlockContainer(shape_type=self._shape)
	
	@property
	def empty(self):
		return self.shape == -1
	
	def set_random_shape(self):
		self.shape = get_random_shape_index()
	

class Box(CtrlElement):
	def __init__(self,height:int,width:int,add_lines:callable):
		self.height 				= height
		self.width  				= width
		self.lines 					= [[CLine.empty for col in range(width)] for l in range(height)]
		self.add_lines 				= add_lines
		self.nextBlock				= BlockContainer()
		self.holdingBlock			= BlockContainer()
		self._can_switch_holding 	= False
		self.nextBlock.set_random_shape()
		super().__init__()
		self.create_new_block()

	def switch_holding(self):
		if not (self.block and self._can_switch_holding):
			return
		
		def get_valid_block(nb:Block):
			for r in [0,1,2,3]:
				for x in [0,-1,1,-2,2]:
					test_block = nb.move(x=x,rotate=r)
					if self.valid_block(test_block):
						return test_block

			return False
		
		nb             : Block = self.block.move()
		nb.shape_index         = get_random_shape_index() if self.holdingBlock.empty else self.holdingBlock.shape
		nb                     = get_valid_block(nb)

		if not nb:
			return
		
		self.holdingBlock.shape = self.block.shape_index
		self.block = nb
		self.changed()
		self._can_switch_holding = False


	def valid_block(self,b:Block):
		for x,y in b.get_positions():
			if not ( (0 <= x < self.width) and (0 <= y < self.height) and (self.lines[y][x] == CLine.empty) ):
				return False
		return True
	
	def get_data(self) -> Data.Box:
		tmp_val = [l[:] for l in self.lines]
		def update(b):
			def get_ghost() -> Block:
				ret_block 	= b.move()
				new_block 	= ret_block.move(y=1)
				while self.valid_block(new_block):
					ret_block = new_block
					new_block = new_block.move(y=1)
				return ret_block
			ghost_block = get_ghost()
			for x,y in ghost_block.get_positions():
				tmp_val[y][x] = CLine.ghost
			for x,y in b.get_positions():
				tmp_val[y][x] = b.shape_index

		shape_index = None
		if self.block:
			shape_index = self.block.shape_index
			update(self.block.move())
		return Data.Box(lines=[[x,y,tmp_val[y][x]] for y,x in product(range(self.height),range(self.width))],shape_type=shape_index)
	
	def create_new_block(self):
		
		b                    = Block(x=math.floor(self.width/2),y=0,shape_index=self.nextBlock.shape)
		if self.valid_block(b):
			self.nextBlock.set_random_shape()
			self.block               = b
			self._can_switch_holding = True
			
		else:

			self.block     = None
			self.game_over = True

		self.changed()
	def new_block(self,b:Block=None,trying_to_move=False):
		if b is None:
			b = Block(x=math.floor(self.width/2),y=0)
		
		if self.valid_block(b):
			self.block = b
			if not trying_to_move:
				self._can_switch_holding = True
			self.changed()
			return b
		
		if trying_to_move:
			return False
		
		self.block     = None
		self.game_over = True
		self.changed()


	def move(self,x=0,y=0,rotate=0,drop_down=False):
		if not self.block:	
			return
		
		def block_dropped():
			for x,y in self.block.get_positions():
				self.lines[y][x] = self.block.shape_index
			new_lines		= list(filter(ar_passes_all(lambda x : x >= 0,as_function=True,inverse=True),self.lines))
			lines_scored 	= len(self.lines) - len(new_lines)
			if lines_scored:
				self.lines = [[CLine.empty for x in range(self.width)] for y in range(lines_scored)] + new_lines
				self.add_lines(lines_scored)

			
			self.create_new_block()
		
		def move_and_set_if_valid(x=0,y=0,rotate=0,ar=[],failed_cb=None,repeat=False):
			ar = [dict(x=x,y=y,rotate=rotate)] if not ar else ar
			def try_ar():
				for t in ar:
					if self.new_block(self.block.move(**t),trying_to_move=True):
						return True
				return False
			
			while try_ar():
				if not repeat:
					return
			
			if failed_cb: 
				failed_cb()
			
		
		if x: 			return move_and_set_if_valid(x=x)
		if y: 			return move_and_set_if_valid(y=y,failed_cb=block_dropped)
		if rotate: 		return move_and_set_if_valid(ar=[dict(x=x,rotate=rotate) for x in [0,-1,1,-2,2]])
		if drop_down: 	return move_and_set_if_valid(y=1,failed_cb=block_dropped,repeat=True)

class Tetris:
	def __init__(self):
		self.height			= 20
		self.width			= 10
		self.score     		= Score()
		self.box    		= Box(self.height,self.width,add_lines=self.score.add_lines)
	


	@property
	def elements(self):
		return dict(box=self.box,score=self.score,nextBlock=self.nextBlock,holdingBlock=self.holdingBlock)
	@property
	def nextBlock(self):
		return self.box.nextBlock

	


	@property
	def holdingBlock(self):
		return self.box.holdingBlock

	


		
		
	@property
	def time_interval(self):
		return ScoreSetting.time_interval_at_level[self.score.level]
	


import curses, curses.ascii, traceback
from curses.textpad 				import Textbox, rectangle
from typing 						import List
from prompt_toolkit.shortcuts 		import input_dialog, message_dialog
from ...fs 							import JDir, JFile
from ...fs.fs 						import JDirFileWrapper
from ...util 						import cb_func
from .. 							import Ctrl, Menu, MenuItem
from .. 							import addon as MenuAddon
from ..addon.addon 					import Addon
from ..menu 						import Button, Ctrl, Curses, InputChar, Menu, MenuItem



class Plugin(Addon):
	pass

class Explorer(Plugin):
	ctrl   : Ctrl 
	menu   : Menu 
	retval : JFile
	running: bool
	def __init__(self,menu_title=None,looking_for:str=None,start_dir=None,filters=[],ex=[],ctrl:Ctrl=None,cb=None):
		super().__init__(ctrl)
		self.start_dir   = JDir.ToJDir(start_dir,JDir.cwd())
		self.filters     = filters
		self.ex          = ex
		self.looking_for = looking_for
		self.menu_title  = menu_title
		self.current_dir = self.start_dir.copyob
		self.retval      = None
		self.ctrl		 = ctrl
		self.running 	 = False
		self.callback    = cb
	
	def display(self):
		self.running = True
		self.start_menu()
		self.build_current_dir()
		self.ctrl.display()

	def get_menu_title(self):
		if self.menu_title:
			return self.menu_title
		if self.looking_for:
			return f"Select file {self.looking_for}"
		return f"Select file"
	
	def start_menu(self):
		if not self.ctrl:
			self.ctrl = Ctrl()
		ctrl = self.ctrl
		MenuAddon.FirstLetter(ctrl)
		MenuAddon.Scrollbar(ctrl)
		MenuAddon.Searchbox(ctrl)
		self.ctrl = ctrl
		self.menu = self.ctrl.create_menu(self.get_menu_title())

		@self.ctrl.on_key("left")
		def on_key_left(ctrl:Ctrl):
			self.current_dir.up()
			self.build_current_dir()

		@self.ctrl.on_key("right")
		def on_key_right(ctrl:Ctrl):
			if ctrl.selected_item and ctrl.selected_item.data.isdir:
				ctrl.do_key("enter")

	def filter(self,el:JDirFileWrapper):
		if el.isdir:
			return True
		for f in self.filters:
			if not f(el):
				return False
		return True

	def build_current_dir(self):
		self.menu.reset()
		menu : Menu = self.menu
		for ob in filter(self.filter,self.current_dir.reader.get(asob=True,deep=False,ex=self.ex)):
			ob : JDirFileWrapper = ob
			@menu.item(text=ob.name,data=ob)
			def ok(mi:MenuItem):
				x : JDirFileWrapper = mi.data
				if x.isdir:
					self.current_dir.fp = x.fp
					self.build_current_dir()
				else:
					def cb(retval:bool):
						before = self.ctrl.active_menu_index
						self.menu.setActive()
						after = self.ctrl.active_menu_index
						if retval:
							self.callback(x)
							self.menu.ctrl.stop()
						else:
							print(f"declined {x.fp}")
					try:
						q = YesNoQuestion("Select this file ?",ctrl=self.ctrl,cb=cb)
					except:
						print(traceback.format_exc())

class LineInput:
	def __init__(self,title:str,text:str,*validator:List[callable],cb:callable=None):
		self.dialog     = input_dialog(title=title,text=text)
		self.validators = validator
		self.cb         = cb
		self.run_dialog()

	@property
	def valid(self):
		for x in self.validators:
			if not x(self.result):
				return False
		return True
	
	def run_dialog(self):
		self.result = self.dialog.run()
		if self.result is None:
			return self.do_callback(cancel=True)
		if self.valid:
			print('valid')
			return self.do_callback()
		self.run_dialog()

	def do_callback(self,cancel=False):
		if not self.cb:
			return
		if cancel:
			self.cb(None)
		else:
			self.cb(self.result)

class SelecterOption:
	def __init__(self,text:str,value:any):
		self.text = text
		self.value = value

class Selecter(Plugin):
	ctrl   : Ctrl 
	menu   : Menu 
	def __init__(self, title:str,*options:List[SelecterOption],ctrl:Ctrl=None,cb:callable=None):
		super().__init__(ctrl)
		self.ctrl  = ctrl if ctrl is not None else Ctrl()
		self.menu = self.ctrl.create_menu(title)

		for op in options:
			@self.menu.item(text=op.text,data=op.value)
			def op_select(mi:MenuItem):
				cb(mi.data)
		
		self.menu.setActive()

class TextInput:
	def __init__(self,title:str,cb:callable,ctrl:Ctrl=None):
		self.title = title
		self.cb    = cb
		self.ctrl  = ctrl if ctrl is not None else Ctrl()
		curses.wrapper(self.display)

	def button_save_clicked(self):
		self.running = False
	
	def display(self,stdscr):
		self.running = True
		cur          = self.ctrl.cur
		max_y        = cur.max_y
		max_x        = cur.max_x
		editwin      = curses.newwin(max_y-6,max_x-2,4,1)
		self.button  = Button("Save",stdscr,self.button_save_clicked,top=0,right=max_x-2)
		rectangle(stdscr, 3, 0, max_y -2, max_x - 1)
		stdscr.refresh()
		box = MyTextbox(editwin,textInput=self)
		while self.running:
			ch = editwin.getch()
			if not ch:
				continue
			if ch == curses.KEY_MOUSE:
				mPos = Curses.get_mouse()
				self.button.handle_click(mPos)
			else:
				inp = InputChar(ch)
				box.proc_key(inp)

		txt = box.gather()
		self.cb(txt)

class MyTextbox(Textbox):
	def __init__(self, *a, textInput : TextInput=None, **b):
		super().__init__(*a,**b)
		self.textInput = textInput
	
	def proc_key(self,inp:InputChar):
		if inp.name == "f2":
			self.textInput.running = False
			return
		if not self.do_command(inp.code):
			self.textInput.running = False
			return
		self.win.refresh()

class YesNoQuestion(Plugin):
	ctrl   : Ctrl 
	menu   : Menu 
	def __init__(self, question:str,ctrl:Ctrl=None,cb:callable=None):
		super().__init__(ctrl)
		self.ctrl  = ctrl if ctrl is not None else Ctrl()
		self.menu = self.ctrl.create_menu(question)
		@self.menu.item(text="Yes",data=True)
		def ok(mi:MenuItem):
			try:
				cb(mi.data)
			except:
				print(traceback.format_exc())
		@self.menu.item(text="No",data=False)
		def ok(mi:MenuItem):
			try:
				cb(mi.data)
			except:
				print(traceback.format_exc())

		self.menu.setActive()


from . import Validator


class InputParameter(IHasReplacement_tab):
	def __init__(self,
		name        : str,            
		validators  : List[Validator.Validator]=[],
		desc        : str              = None,
		ex          : List[str]        = [],
		start_dir   : str              = None,
		max_size    : int              = None,
		min_size    : int              = None,
		min_len     : int              = None,
		max_len     : int              = None
	):
		self.name       = name
		self.validators = validators
		self.desc       = desc
		self.ex         = ex
		self.start_dir  = start_dir
		self.max_size   = max_size
		self.min_size   = min_size
		self.max_len    = max_len
		self.min_len    = min_len

	@property
	def replacement_tab(self):
		return dict(
		name       = self.name,
		desc       = self.desc)


from ..menu import Menu,MenuItem,Ctrl,addon
from .tester import Tester


class MenuTester(Tester):
	def test(self):
		def test_one():
			menuCtrl = Ctrl(max_lines=29)
			print(menuCtrl.cur.color.normal)
			addon.Searcher(menuCtrl)
			addon.FirstLetter(menuCtrl)
			addon.Scrollbar(menuCtrl)
			addon.Searchbox(menuCtrl)
			mainMain = menuCtrl.create_menu("Test")
			names = ["Joris","Lotte","Kevin","Davy","Ronny","Anke","Aars","Jan","Loes","Lies","Lore"]


			for i in range(2000):
				@mainMain.item(f"testing {i}",data=i)
				def on_click(mi:MenuItem):
					print(mi.data)
					mi.ctrl.stop()

			for n in names:
				@mainMain.item(n,data=n)
				def test_func(mi:MenuItem):
					print(f"OK {mi.data}")
					mi.ctrl.stop()
			menuCtrl.display()

		return [test_one]






from __future__ import annotations
from typing import List
from jivi.util.ob import copy_tab
import re
from jivi.fs import JFile
from ..models import IHasReplacement_tab,string_replacer

class Validator:
	_error_text = "[value] is invalid!"
	parent     : IHasReplacement_tab
	validate   : callable
	value	   : any
	def __init__(self,parent:IHasReplacement_tab):
		self.parent = parent

	@property
	def error_text(self):
		return string_replacer(self._error_text,self.parent.replacement_tab,value=self.value)

	def __repr__(self):
		return f"[<{self.__class__.__name__}>] {self.value}"


def __creater(error_text:str):
	def decorator(validate_func) -> Validator:
		class ret_cls(Validator):
			_error_text = error_text
			def validate(self,value):
				self.value = value
				return validate_func(self,value)
		ret_cls.__name__ = validate_func.__name__
		return ret_cls
	return decorator

class StringNotEmpty(Validator):
	_error_text = "[value] is required!"
	def validate(self,value:str):
		return len(value.strip()) > 0
class Email(Validator):
	_error_text = "[value] is not a valid email!"
	def validate(self,value):
		return re.match("(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)",value)

class FileExists(Validator):
	_error_text = "File [value] does not exist!"
	def validate(self,value):
		return JFile(value).exists
	
class FileHasContent(Validator):
	_error_text = "File [value] has no content or does not exist!"
	def validate(self,value):
		return JFile(value).reader.text

class FileIsJson(Validator):
	_error_text = "File [value] is not a valid json or cant be read!"
	def validate(self,value):
		return JFile(value).reader.json()


@__creater("Is not bigger than 10")
def BiggerThan10(self,value) -> Validator:
	return value > 10




def tester():

	class testparent(IHasReplacement_tab):
		@property
		def replacement_tab(self):
			return dict(name="OMO")
		
	x = BiggerThan10(testparent())

	x.validate(8)
	print(x)
	print(x.error_text)


from ..util import String
import traceback

class Tester:
	
	@property
	def classname(self):
		return self.__class__.__name__
	
	@classmethod
	def do(cls):
		c = cls()
		classname = c.classname
		retvals = []

		for testsection in filter(lambda k : not k in dir(Tester), dir(cls)):
			for f in getattr(c,testsection)():
				retval_el = [classname,testsection,f.__name__]
				try:
					resp = f()
					if resp is not None:
						retval_el += ["Failed",resp]
					else:
						retval_el += ["Succes",""]

					retvals.append(retval_el)
				except:
					print(traceback.format_exc())
					retval_el += ["Failed",traceback.format_exc()]

		for a in String.same_len_ar(retvals):
			print(a)

from time import time

class PerformanceTest:
	def __init__(self,*funcs):
		self.funcs = funcs
	def run(self,runs=500,*a,**b):
		ar = []
		for f in self.funcs:
			ar_el = [f.__name__]
			s = time()
			for i in range(runs):
				f(*a,**b)
			
			ar_el.append(time() - s)
			ar.append(ar_el)

		for a in String.same_len_ar(ar):
			print(a)




from .util 	import Ob,String,Index
from .util 	import Popen,to_int,to_string,good_list_index,all_true,first_item,first_index,cb_func,download_html,convert_func,Thread,is_str_list
from .util  import indexes,indexes_reversed,enumerate_reversed
from .util	import Pause,Clearterminal,random_el,enumerate_random,index_random

from .util import ar_flatten,export


# failed ,kw_or_self,ob_set


from .models import copy_tab
from .inputparameters import InputParameter
from . import ValueGetter
from .base import UIInput as base


class UIInput(base.UIInput):
	para : InputParameter
	_method_classes : List[callable] = []
	def __init__(self,para : InputParameter):
		self.para = para

	@property
	def my_replacement_tab(self):
		return dict()
	
	@property
	def replacement_tab(self):
		tab = copy_tab(self.para.replacement_tab)
		tab.update(self.my_replacement_tab)
		return tab
	
	@property
	def inputMethods(self):
		return [cls(self) for cls in self._method_classes]
	
	def __repr__(self):
		return f"[<{self.__class__.__name__}>] {self.para}"

class File(base.File):
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter]

class Json(base.Json):
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter,ValueGetter.Text]

class Text(base.Text):
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter,ValueGetter.Text]

class String(base.String):
	_method_classes = [ValueGetter.String]

class Number(base.Number):
	_method_classes = [ValueGetter.Number]


import traceback
from time import sleep, time
from typing import List

from ..constant import TetrisCtrlEvent,DisplayData
from jivi.thread import Thread

from ..controller import CtrlElement
from ..controller import Tetris as Controller
from ..view import View




class Tetris:
	ctrlTrg = TetrisCtrlEvent
	view : View
	ctrl : Controller
	dispData : DisplayData
	def __init__(self,viewClass:View):
		self.ctrl = Controller()

		self.dispData = DisplayData({k : v.get_data for k,v in self.ctrl.elements.items()})
		self.view = viewClass(self.dispData)
		self._running_index = 0
		self.register_listeners()
		

	def start(self):
		self.view.start()
		self.running = True
		self._running_index += 1
		def thread_condition(running_index):
			return self.running and self._running_index == running_index
		
		def thread_first_run(*a,**b):
			self.next_auto_drop = time() + self.ctrl.time_interval
		@Thread.register(options=Thread.options(interval=0.1,condition=thread_condition,first_run=thread_first_run),args=[self._running_index])
		def tick_thread(running_index):
			if time() > self.next_auto_drop:
				self.ctrl.box.move(y=1)
				self.next_auto_drop = time() + self.ctrl.time_interval

			self.dispData.lastChanged = {k : v.last_change for k,v in self.ctrl.elements.items()}

			self.view.check_changes()
			


		while self.running:
			sleep(0.1)
		self._running_index += 1
	def register_listeners(self):
		cBox = self.ctrl.box
		c = self.ctrlTrg
		@c.do_exit()
		def f(ev):
			self.exit()

		@c.left()
		def f(ev):
			cBox.move(x=-1)
		
		@c.right()
		def f(ev):
			cBox.move(x=+1)
		
		@c.down()
		def f(ev):
			cBox.move(y=1)
		
		@c.up()
		def f(ev):
			cBox.move(rotate=1)
		
		@c.holding()
		def f(ev):
			self.ctrl.box.switch_holding()
		
		@c.drop()
		def f(ev):
			cBox.move(drop_down=True)
		
	def exit(self):
		self._running_index += 1
		self.running = False
		self.view.exit()
		Thread.stop_all()



from __future__ 			import annotations
from typing import List
from ..constant import CLine,GHOST_BLOCK_CHAR,ALL_SHAPES,Data
from jivi.Curses import COLOR,get_keyname,valid_key,rectangle,curses
from jivi.event import EventWithKey
from ..constant import TetrisCtrlEvent,DisplayData
from time import time
from jivi.thread import Thread


class ViewElement:
	def __init__(self):
		self.init()
		self.last_update = 0

	def init(self):
		pass

	def _draw_dynamic(self,data):
		pass

	def draw_dynamic(self,data,last_update):
		self.last_update = last_update
		self._draw_dynamic(data)
	


class HoldingBlock(ViewElement):
	dataType = Data.BlockContainer



class NextBlock(ViewElement):
	dataType = Data.BlockContainer


class Score(ViewElement):
	dataType = Data.Score

class Box(ViewElement):
	last_lines = None
	def _draw_dynamic(self,data:Data.Box):
		BLOCK_TYPE_INDEX = 2

		def get_diference():
			if not self.last_lines:
				for a in data.lines:
					yield a
			else:
				for index in range(len(data.lines)):
					if self.last_lines[index][BLOCK_TYPE_INDEX] != data.lines[index][BLOCK_TYPE_INDEX]:
						yield data.lines[index]

			self.last_lines = data.lines
		
		self._draw_diferences(get_diference(),data)

	def _draw_diferences(self,difference,data:Data.Box):
		pass

	




class Tetris:
	ctrlTrg = TetrisCtrlEvent
	nextBlock    : NextBlock   
	holdingBlock : HoldingBlock
	box     : Box         
	score   : Score       
	def __init__(self,dispData:DisplayData):
		self.running = False
		self.dispData = dispData
		self.init()

	def init(self):
		pass
	def check_changes(self):
		pass

	@property
	def elements(self):
		return dict(box=self.box,score=self.score,nextBlock=self.nextBlock,holdingBlock=self.holdingBlock)
class BaseView:
	Tetris       = Tetris
	Score        = Score
	NextBlock    = NextBlock
	HoldingBlock = HoldingBlock
	Box          = Box
	ViewElement  = ViewElement


from . 			import util 		as 	jvUtil
from .wrapper 	import 					jvWrapper,jvWrapperSuper
from . 			import menu 		as 	jvMenu
from . 			import fs 			as 	jvFs
from .          import event        as  jvEvent



class jv:
	menu    = jvMenu
	wrapper = jvWrapper
	util    = jvUtil
	fs      = jvFs
	event   = jvEvent


from typing import List,Dict
from typing import TypeVar, Generic
from logging import Logger

T = TypeVar('T')

class CBOptions:
	clear_args : bool     	= False
	add_args   : List[any]	= []
	add_kwargs : dict     	= dict()
	condition  : callable 	= None
	no_error   : bool     	= False
	priority   : int		= 2
	def __init__(self,
				priority 			   = 2,
				clear_args             = False,
				add_args               = [],
				add_kwargs             = dict(),
				condition              = None,
				no_error			   = False):
		self.set(priority=priority,clear_args=clear_args,add_args=add_args,add_kwargs=add_kwargs,condition=condition,no_error=no_error)
	
	def set(self,priority=priority,clear_args= False,add_args= [],add_kwargs= dict(),condition=None,no_error=False):
		self.priority   = priority
		self.clear_args = clear_args
		self.add_args   = add_args
		self.add_kwargs = add_kwargs
		self.condition  = condition
		self.no_error   = no_error
		return self
	
	@property
	def copy(self):
		return self.__class__(priority=self.priority,clear_args=self.clear_args,add_args=self.add_args,add_kwargs=self.add_kwargs,condition=self.condition,no_error=self.no_error)


class CB:
	def __init__(self,func : callable,options:CBOptions=CBOptions()):
		self.original_func = func
		self.func          = self.original_func
		self.options       = options
		if options.condition:
			self.__convert_condition(options.condition)
		if options.no_error:
			self.__convert_no_error()
	
	def __convert_condition(self,condition):
		f = self.func
		def new_func(*a,**b):
			if condition(*a,**b):
				return f(*a,**b)
		self.func = new_func

	def __convert_no_error(self):
		f = self.func
		def new_func(*a,**b):
			try:
				retval = f(*a,**b)
				return retval
			except:
				pass
		self.func = new_func
	
	def __call__(self,*a,**b):
		if self.options.clear_args:
			return self.func(*self.options.add_args,**self.options.add_kwargs)
		else:
			for k,v in self.options.add_kwargs.items():
				b[k] = v
			return self.func(*a,*self.options.add_args,**b)

			
class Event:
	__callbacks : List[CB]
	_is_single  : bool     = False
	_running    : bool     = False

	def __init__(self):
		self.__callbacks = []
	def before_fire(self,*a,**b):
		pass

	def fire(self,*a,**b):
		self._running = True
		self.before_fire(*a,**b)
		for cb in sorted(self.__callbacks,key=lambda x : x.options.priority):
			if self._running:
				cb(self)
			else:
				break
	
	
	def stopPropagation(self):
		self._running = False
	

	@property
	def cb(self):
		return CBOptions().set
	
	def register(self,options:CBOptions=CBOptions(),func=None):
		def add_func(f):
			cb = CB(f,options)
			if self._is_single:
				self.__callbacks = [cb]
			else:
				self.__callbacks.append(cb)
			return f
		if func:
			return add_func(func)
		
		return add_func

class SingleEvent(Event):
	_is_single   : bool     = True



class EventWithKey:
	__keys     : Dict[str,Event] = dict()
	eventClass : Event           = Event
	_is_single  : bool    		 = True
	def __init__(self,eventClass=Event,format_key=None,valid_key=None):
		self.__keys               = dict()
		self.eventClass           = eventClass
		self.eventClass._is_single = self._is_single
		if format_key:
			self.format_key = format_key
		if valid_key:
			self.valid_key = valid_key
		
	@property
	def cb(self):
		return CBOptions().set
	
	def format_key(self,inp):
		return inp
	
	def valid_key(self,key):
		return True

	def get(self,key):
		key = self.format_key(key)
	
		if key in self.__keys:
			return self.__keys[key]
		if self.valid_key(key):
			self.__keys[key] = self.eventClass()
			return self.__keys[key]

	def fire(self,key,*a,**b):
		x = self.get(key)
		if not x:
			return
		x.fire(*a,**b)

	def register(self,key,options:CBOptions=CBOptions(),func=None):
		if ev := self.get(key):
			return ev(options,func=func)



class SingleEventWithKey(EventWithKey):
	_is_single = True





def kw_or_self(f:callable):
	named_arguments = list(f.__code__.co_varnames[len(f.__code__.co_varnames)-len(f.__defaults__):])
	def retval(self,*a,**b):
		for k in filter(lambda x : (b.get(x,None) is None) and hasattr(self,x), named_arguments):
			b[k] = getattr(self,k)
		f(self,*a,**b)
	return retval


def key_list_not_null(t):
	return ','.join([k for k,v in t.items() if v is not None])

def ob_set(f:callable=None,kw_or_self=None):
	
	if f is not None:
		named_arguments = {k : f.__defaults__[ix] for ix,k in enumerate(f.__code__.co_varnames[len(f.__code__.co_varnames)-len(f.__defaults__):])}
		fstr = str(f)

		do_print = fstr.find("Box") != -1
		if do_print:
			print("varnames",f.__code__.co_varnames)
			print("varnames len",len(f.__code__.co_varnames))
			print("defaults len",len(f.__defaults__))
			print("defaults",f.__defaults__)
			print("named_arguments",named_arguments)
			print("named_arguments len",len(named_arguments))
		def retval(self,*a,**b):
			if do_print:
				print(self.__class__.__name__,named_arguments)
			#print("b",key_list_not_null(b))
			#print(f.__name__,self,a,b)
			if kw_or_self:
				for k in filter(lambda x : (b.get(x,None) is None) and hasattr(self,x), named_arguments):
					b[k] = getattr(self,k)
			bb = {k : b.get(k,v) for k,v in named_arguments.items()}
			if do_print:
				print(self.__class__.__name__,"bb",bb)
			for k,v in bb.items():
				setattr(self,k,v)
			f(self,*a,**b)
		return retval
	else:
		def retval(f,kw_or_self=kw_or_self):
			return ob_set(f,kw_or_self=kw_or_self)
		return retval

def ob_init_props(ob,locs,not_none=True):
	for k,v in locs.items():
		if k == "self":
			continue
		if not_none and v is None:
			continue
	
		setattr(ob,k,v)



import math
from typing import List
from jivi.event import Event

def get_all_shapes():
	def get_rotations(base_shape):
		def to_pos_ar(ar):
			ret_ar = []
			start_x = 0 - math.ceil(len(ar[0])/2)

			for y,line in enumerate(ar):
				for x,val in enumerate(line):
					if not val:
						continue

					ret_ar.append([x+start_x,y])
			return ret_ar


		def mirrorv(ar):
			h = len(ar)
			w = len(ar[0])
			return [[ar[i][w - 1 - j] for j in range(w)] for i in range(h)]


		def mirrorh(ar):

			h = len(ar)
			w = len(ar[0])
			return [[ar[h - 1 - i][j] for j in range(w)] for i in range(h)]
		
		def transpose(ar):
			h = len(ar)
			w = len(ar[0])
			return [[ar[i][j] for i in range(h)] for j in range(w)]
		ar = [[int(c) for c in l] for l in base_shape]
		
		rotations = []

		for funcar in [[],[transpose,mirrorv],[mirrorh,mirrorv],[transpose,mirrorh]]:
			rotation = ar
			for f in funcar:
				rotation = f(rotation)
			rotations.append(rotation)
	
		return [to_pos_ar(x) for x in rotations]
	ret_val = []
	def a(base_shape):
		ret_val.append(get_rotations(base_shape))

	a([
		"111",
		"010"
	])
	a([
		"011",
		"110"
	])
	a([
		"110",
		"011"
	])
	a([
		"1111"
	])
	a([
		"11",
		"11"
	])
	a([
		"111",
		"100"
	])
	a([
		"111",
		"001"
	])
	return ret_val

ALL_SHAPES = get_all_shapes()




LEVEL_SCORE = [0,100,200,500,1000,2000,10000,30000,50000,100000]

def get_level_scores_till_100():
	retval = []
	FINAL_LEVEL_SCORE = LEVEL_SCORE[-1]
	LEVEL_SCORE_NOW = 0
	for i in range(100):
		if i < len(LEVEL_SCORE):
			LEVEL_SCORE_NOW += LEVEL_SCORE[i]
		else:
			LEVEL_SCORE_NOW += FINAL_LEVEL_SCORE
		retval.append(LEVEL_SCORE_NOW)
	return retval

def get_time_interval_at_level_till_100():
	retval = [1,0.9,0.8,0.7,0.6,0.5,0.4,0.3,0.2]
	retval += [0.2]*(100 - len(retval))
	return retval



class ScoreSetting:
	score_for_lines_made 	= [0,5,10,15,100]
	score_at_level       	= get_level_scores_till_100()
	time_interval_at_level	= get_time_interval_at_level_till_100()


GHOST_BLOCK_CHAR = bytes((178,)).decode('cp437')

class CLine:
	empty = -1
	ghost = -2



	

class Data:
	class Box:
		def __init__(self,lines:List[List[int]],shape_type:int=None):
			self.lines      = lines
			self.shape_type = shape_type

		def to_dict(self):
			return dict(lines=self.lines,shape_type=self.shape_type)
	class Score:
		def __init__(self,level:int=0,score:int=0,lines:int=0):
			self.level = level
			self.score = score
			self.lines = lines
		def to_dict(self):
			return {k : getattr(self,k) for k in 'level,score,lines'.split(',')}
	class BlockContainer:
		def __init__(self,shape_type:int=None):
			self.shape_type = shape_type


		def to_dict(self):
			return dict(shape_type=self.shape_type)


class DisplayData:
	def __init__(self,get_data_dict):
		self.lastChanged = dict(box=0,score=0,nextBlock=0,holdingBlock=0)
		self.getData     = get_data_dict


class TetrisCtrlEventClass:
	do_exit : Event
	left    : Event
	right   : Event
	down    : Event
	up      : Event
	holding : Event
	drop    : Event
	def __init__(self):
		self.do_exit = Event()
		self.left    = Event()
		self.right   = Event()
		self.down    = Event()
		self.up      = Event()
		self.holding = Event()
		self.drop    = Event()

	__singleton = None
	@classmethod
	def get(cls):
		if not cls.__singleton:
			cls.__singleton = cls()
		return cls.__singleton
TetrisCtrlEvent : TetrisCtrlEventClass = TetrisCtrlEventClass.get()


import socket,traceback
from jivi.thread import Thread
import socketserver
import traceback
import json


class MyTCPHandler(socketserver.BaseRequestHandler):
	def handle(self):
		self.data = self.request.recv(1024).strip()
		print("{} wrote:".format(self.client_address[0]))
		print(self.data)

class TCPServer:
	server : socketserver.TCPServer
	def __init__(self,host:str="localhost",port:int=None,cb:callable=None,decode=None,is_json=False):
		self.host    = host
		self.port    = port
		self.cb      = cb
		self.decode  = decode
		self.is_json = is_json
		self.server  = None
		Thread.register(options=Thread.options(no_loop=True),func=self.create_server)


	def data_handler(self,rawdata):
		data = rawdata
		try:
			if self.decode:
				data = data.decode(self.decode)
			if self.is_json:
				data = json.loads(data)

			self.cb(data)
		except:
			print(traceback.format_exc())
		
	def shutdown(self):
		if self.server:
			self.server.shutdown()
		
	def create_server(self):
		data_handler = self.data_handler
		class MyTCPHandler(socketserver.BaseRequestHandler):
			def handle(self):
				data_handler(self.request.recv(1024).strip())
		with socketserver.TCPServer((self.host, self.port), MyTCPHandler) as server:
			self.server = server
			server.serve_forever()


class TCPClient:
	@staticmethod
	def send(port:int,host:str="localhost",msg:str="No message",cb=None):
		
		try:
			msg_byte = msg.encode()
			s        = socket.socket()
			s.connect((host,port))
			s.send(msg_byte)
			s.close()
			return True
		except:
			#print(traceback.format_exc())
			return False




from .tester import Tester
from ..uinput import UIInput,Validator,ValueGetter,InputParameter


class UIInputTester(Tester):
	def test(self):
		def create_and_print():
			x = UIInput.String(para=InputParameter(name="Email",validators=[Validator.StringNotEmpty,Validator.Email]))
			for inputMethod in x.inputMethods:
				print(inputMethod)

		return [create_and_print]

	


from __future__		import annotations
from typing 		import List
from curses.textpad import rectangle
from jivi.util 		import random_el,Thread,index_random
from jivi.Curses 	import COLOR,CURSES,InputChar,Position,CursesWrapper
import curses
from time import sleep

GHOST_BLOCK = bytes((178,)).decode('cp437')
GHOST_BLOCK = GHOST_BLOCK + GHOST_BLOCK
FREE_SPOT = -1
GHOST_SPOT = -2


LINE_SCORE = [0,5,10,15,100]
LEVEL_SCORE = [0,100,200,500,1000,2000,10000,30000,50000,100000]

def __get_level_scores_till_100():
	retval = []
	FINAL_LEVEL_SCORE = LEVEL_SCORE[-1]
	LEVEL_SCORE_NOW = 0
	for i in range(100):
		if i < len(LEVEL_SCORE):
			LEVEL_SCORE_NOW += LEVEL_SCORE[i]
		else:
			LEVEL_SCORE_NOW += FINAL_LEVEL_SCORE
		retval.append(LEVEL_SCORE_NOW)
	return retval

SCORE_AT_LEVEL = __get_level_scores_till_100()

class Shapes:
	ZShape     = [[0,-1],[0,0],[-1,0],[-1,1]]
	SShape     = [[0,-1],[0,0],[1,0],[1,1]]
	IShape     = [[0,-1],[0,0],[0,1],[0,2]]
	TShape     = [[-1,0],[0,0],[1,0],[0,1]]
	Square     = [[0,0],[1,0],[0,1],[1,1]]
	LShape     = [[-1,-1],[0,-1],[0,0],[0,1]]
	JShape     = [[1,-1],[0,-1],[0,0],[0,1]]
	ALL        = [ZShape,SShape,IShape,TShape,Square,LShape,JShape] 

	@classmethod
	def random(cls):
		i = index_random(cls.ALL)
		return COLOR.B.ALL[i],cls.ALL[i]
class Element:
	curs          : CursesWrapper   = None
	@property
	def total_width(self) -> int:
		return self.offset_x + self.width
	
	@property
	def total_height(self) -> int:
		return self.offset_y + self.height
	
	@property
	def win(self) -> curses.window:
		return self.curs.win

def line_is_done(ar):
	for x in ar:
		if x == FREE_SPOT:
			return False
	return True

class Block:
	shape_index : int       = None
	y           : int       = 1
	x           : int       = 5
	box         : Box       = None
	shape       : List[int] = None
	def __init__(self,box:Box=None,shape_index:int=None,shape:List[int]=None,x:int=5,y:int=1):
		self.x 				= x
		self.y 				= y
		self.box 			= box
		self.shape_index 	= shape_index 	if shape_index is not None 		else index_random(Shapes.ALL)
		self.shape 			= shape 		if shape is not None			else Shapes.ALL[self.shape_index]
		self.grounded		= False





	def can_move(self,x,y,shape=None):
		return self.box.free_pos_shape(x,y,self.shape if shape is None else shape)

	def move(self,x=None,y=None,rotate=None,rotate_inverse=None):
		if x:
			if self.can_move(self.x+x,self.y):
				self.x = self.x+x
				self.box.draw()
				return True
		elif y:
			target_y = self.y + y
			while self.y != target_y:
				if self.can_move(self.x,self.y+1):
					self.y   += 1
				else:
					return self.box.set_new_block(self)
			self.box.draw()
		elif rotate or rotate_inverse:
			new_shape = [[a[1],-a[0]][:] for a in self.shape] if rotate else [[-a[1],a[0]][:] for a in self.shape]
			for x_delta in [0,-1,-2,1,2]:
				new_x = self.x + x_delta
				if self.can_move(new_x,self.y,new_shape):
					self.shape = new_shape
					self.x = new_x
					self.box.draw()
					return
	
	@property
	def ghost_positions(self):
		def gety():
			ghost_y  = self.y
			while True:
				if self.can_move(self.x,ghost_y+1):
					ghost_y += 1
				else:
					return ghost_y
		ghost_y = gety()
		return [[self.x + x,ghost_y + y] for x,y in self.shape]
	
	@property
	def positions(self):
		return [[self.x + x,self.y + y] for x,y in self.shape]

	@property
	def color_dict(self,show_ghost=True):
		retval = dict()
		if show_ghost:
			for x,y in self.ghost_positions:
				if not y in retval:
					retval[y] = dict()
				retval[y][x] = GHOST_SPOT
		for x,y in self.positions:
			if not y in retval:
				retval[y] = dict()
			retval[y][x] = self.shape_index
		return retval
	
	@property
	def color(self):
		return COLOR.B.ALL[self.shape_index]

class Box(Element):
	block         : Block            = None
	lines         : List[List[int]] = []
	old_color     : List[List[int]] = []
	offset_x      : int              = 1
	offset_y      : int              = 0
	height        : int              = 20
	block_per_row : int              = 10
	curs          : CursesWrapper   = None

	def __init__(self,curs:Tetris=None,offset_x=None,offset_y=None,block_per_row=None,height=None,show_ghost=True):
		self.__event_line_scored_listeners 	= []
		self.curs   						= curs
		self.offset_x 						= offset_x 		if offset_x			is not None else self.offset_x
		self.offset_y 						= offset_y 		if offset_y			is not None else self.offset_y
		self.block_per_row					= block_per_row if block_per_row 	is not None else self.block_per_row
		self.height							= height		if height			is not None else self.height
		self.lines  						= [[-1 for x in range(self.block_per_row)] for y in range(self.height)]
		self.show_ghost						= show_ghost
		self.draw_static()
		self.set_new_block()
		self.register_keys()
		self.draw()

	def event_line_scored_register(self,func):
		self.__event_line_scored_listeners.append(func)
	
	def set_new_block(self,block:Block=None):
		if block:
			for x,y in block.positions:
				self.lines[y][x] = block.shape_index

		new_lines = []
		for l in self.lines:
			if not line_is_done(l):
				new_lines.append(l)
		
		lines_scored = len(self.lines) - len(new_lines)
		if lines_scored:
			
			for f in self.__event_line_scored_listeners:
				f(lines_scored)
			
			new_lines = [[-1 for x in range(self.block_per_row)] for y in range(lines_scored)] + new_lines
			self.lines = new_lines
		
		self.block = Block(box=self)
		self.draw()
	
	def draw_static(self):
		rectangle(self.win,0,0,1+self.height,self.offset_x+self.width)
	
	def draw_old(self):
		color_conv = dict()
		color_conv[-1] = COLOR.B.BLACK
		color_conv[-2] = COLOR.B.MAGENTA

		block_color_dict = self.block.color_ghost_dict if self.show_ghost else dict
		for y,t in self.block.color_dict.items():
			for x,color in t.items():
				block_color_dict[y][x] = color
			
		new_color = []
		for y,color_ar in enumerate(self.lines):
			block_row_dict = block_color_dict.get(y,dict())
			for x,color in enumerate(color_ar):
				color_value = block_row_dict.get(x,color)
				if color_value in color_conv:
					color_value = color_conv[color_value]
				new_color.append([y,x,color_value])

		if self.old_color:
			differences = []
			for i in range(len(new_color)):
				if new_color[i][2] != self.old_color[i][2]:
					differences.append(new_color[i])
		else:
			differences = new_color

		for y,x,color in differences:
			self.win.addstr(y+1,x*2+self.offset_x,"  ",color)
	
	def draw(self):
		color_conv = dict()
		color_conv[-1] = COLOR.B.BLACK   #EMPTY
		color_conv[-2] = COLOR.B.MAGENTA #GHOST
		char_conv = dict()
		char_conv[-1] = "  "
		char_conv[-2] = GHOST_BLOCK
		for i,c in enumerate(COLOR.B.ALL):
			char_conv[i] = "  "
			color_conv[i] = c
		

		block_color_dict = self.block.color_dict

		new_color = []
		for y,color_ar in enumerate(self.lines):
			block_row_dict = block_color_dict.get(y,dict())
			for x,block_type_index in enumerate(color_ar):
				block_type_index = block_row_dict.get(x,block_type_index)
				new_color.append([y,x,block_type_index])

		if self.old_color:
			differences = []
			for i in range(len(new_color)):
				if new_color[i][2] != self.old_color[i][2]:
					differences.append(new_color[i])
		else:
			differences = new_color

		for y,x,block_type_index in differences:
	
			a = char_conv[block_type_index]
			b = color_conv[block_type_index]
			self.win.addstr(y+1,x*2+self.offset_x,char_conv[block_type_index],color_conv[block_type_index])


		self.win.refresh()


	def free_pos_shape(self,x:int,y:int,shape:List[List[int]]):
		for x_shape,y_shape in shape:
			X,Y = x+x_shape,y+y_shape
			if ((X < 0 or X >= self.block_per_row) or (Y < 0 or Y >= self.height) or (self.lines[Y][X] != FREE_SPOT)):
				return False
		return True
	
	def register_keys(self):
		def key(key,func,*a,no_arg_default=True,**b):
			b["no_arg_default"] 	= no_arg_default
			self.curs.on_key(key,func,**b)
		key('space',self.block_move,kwargs=dict(y=100))
		key('down',self.block_move,kwargs=dict(y=1))
		key('right',self.block_move,kwargs=dict(x=1))
		key('left',self.block_move,kwargs=dict(x=-1))
		key('up',self.block_move,kwargs=dict(rotate_inverse=True))

	def block_move(self,x=None,y=None,rotate=None,rotate_inverse=None):
		self.block.move(x=x,y=y,rotate=rotate,rotate_inverse=rotate_inverse)

	@property
	def width(self) -> int:
		return self.block_per_row*2

class Score(Element):
	curs     : CursesWrapper = None
	offset_x : int           = 1
	offset_y : int           = 0
	width    : int           = 20
	height   : int           = 5
	__level  : int           = 0
	__score  : int           = 0
	__lines  : int           = 0
	

	@property
	def time_interval(self):
		return 1

	@property
	def score(self):
		return self.__score
	
	@score.setter
	def score(self,value):
		self.__score = value

		next_level = self.level
		while value > SCORE_AT_LEVEL[next_level+1]:
			next_level += 1
		
		if next_level != self.level:
			self.level = next_level
	
	@property
	def level(self):
		return self.__level
	@level.setter
	def level(self,value):
		self.__level = value
		self.draw()
	
	@property
	def lines(self):
		return self.__lines
	@lines.setter
	def lines(self,value):
		to_add_value = value - self.__lines
		self.__lines = value
		self.score += LINE_SCORE[to_add_value]
		self.draw()

	def __init__(self,curs:CursesWrapper=None,offset_x=None,offset_y=None,height=None,width=None,score=0,level=0,lines=0):
		self.curs   		= curs
		self.offset_x 		= offset_x 		if offset_x			is not None else self.offset_x
		self.offset_y 		= offset_y 		if offset_y			is not None else self.offset_y
		self.width			= width 		if width		 	is not None else self.width
		self.height			= height		if height			is not None else self.height
		self.__level 		= level
		self.__score		= score
		self.__lines		= lines
		self.draw_static()
	
	def draw_static(self):
		def add(y,x,s):
			self.win.addstr(self.offset_y+y,self.offset_x+x,s,COLOR.F.WHITE)

		rectangle(self.win,self.offset_y,self.offset_x,self.offset_y+self.height,self.offset_x+self.width)
		add(1,1,"Level")
		add(2,1,"Lines")
		add(3,1,"Score")
	
	def draw(self):
		end_x = self.offset_x + self.width
		def add(y,n):
			s  = str(n)
			sl = len(s)
			x  = end_x-sl
			self.win.addstr(self.offset_y+y,x,s,COLOR.F.WHITE)
		add(1,self.level)
		add(2,self.lines)
		add(3,self.score)

	def line_scored(self,amount):
		self.lines += amount

class Tetris(CursesWrapper):
	score      : Score = None
	show_lines : bool  = False
	show_ghost : bool  = False
	box        : Box   = None
	def __init__(self,show_lines=False,show_ghost=True):
		super().__init__()
		self.show_lines = show_lines
		self.show_ghost = show_ghost
		curses.wrapper(self.initscr)

	def initscr(self,win:curses.window):
		super().initscr(win)
		on_key     = self.on_key
		@on_key('esc')
		def f(m:Tetris):
			m.stop()
		
		self.box   = Box(curs=self,show_ghost=self.show_ghost)
		self.score = Score(curs=self,offset_x=self.box.total_width+1)
		self.box.event_line_scored_register(self.score.line_scored)
		self.key_reader()

if __name__ == "__main__":
	Tetris(show_lines=False)


from __future__ 		import annotations


def get_all():
	from .download 			import DownloadTester
	from .fs				import FSTester
	from .menu				import MenuTester
	from .util				import UtilTester
	from .wrapper			import WrapperTester
	from .menu_plugin   	import PluginTester

	return {k : v for k,v in locals().items()}






class Test:
	def __init__(self):
		from .tester import Tester
		from ..util  import Pause,Clearterminal
		current_tester : Tester = None
		from ..menu.menu 	import Ctrl
		from ..menu.plugin 	import YesNoQuestion
		ctrl = Ctrl()
		def cb(val):
			if val:
				ctrl.exit()
				current_tester.do()
				Pause()
			else:
				pass
			ctrl.stop()
	

		for k,v in get_all().items():
			current_tester = v
			YesNoQuestion(f"Test {k} ?",ctrl,cb=cb)
			ctrl.display()
		ctrl.exit()
	@classmethod
	def optionMenu(cls):
		from .tester import Tester
		from ..util  import Pause,Clearterminal
		current_tester : Tester = None
		from ..menu.menu 	import Ctrl,MenuItem
		ctrl = Ctrl()

		@ctrl.on_key('esc')
		def f(*a,**b):
			ctrl.exit()
			
		mainMenu = ctrl.create_menu("Select item to test")

		for k,v in get_all().items():
			@mainMenu.item(k,data=v)
			def on_click(mi:MenuItem):
				ctrl.cur.stop()
				mi.data.do()
				Pause()
		ctrl.display()

		ctrl.stop()
		ctrl.exit()



from __future__ 			import annotations
from typing 				import List
from ..util  				import first_index,first_item,good_list_index,all_true,convert_func,String
from curses.textpad 		import rectangle
from time 					import sleep,time
import curses
import math


class Event:

	class Event:
		debug = False
		__registered_funcs : List[callable]
		def __init__(self,ctrl):
			self.ctrl = ctrl
			self.__registered_funcs = []
		
		@property
		def menu(self):
			return self.ctrl.active_menu
		
		def before_fire(self,*a,**b):
			pass
		def fire(self,*a,**b):
			self.before_fire(*a,**b)
			for f in self.__registered_funcs:

				f(self)
		
		def __call__(self,condition=None):
			def decorator(f):
				if condition is not None:
					def new_func(*a,**b):
						if condition(*a,**b):
							f(*a,**b)
					self.register_func(new_func)
					return new_func
				self.register_func(f)
				return f
			return decorator
		def init(self):
			pass

		def register_func(self,func):
			self.__registered_funcs.append(func)
	
	class MouseClick(Event):
		coord        : Position
		def before_fire(self,*a,**b):
			id, mx, my, z, bstate= curses.getmouse()
			self.coord = Position(mx,my)

	class CharReceived(Event):
		received:InputChar = None
		def before_fire(self,received:InputChar,*a,**b):
			self.received = received

	class FocusedOnListChanged(Event):
		focused_now:bool=False
		def before_fire(self,focused_now,*a,**b):
			self.focused_now = focused_now

class Events:
	mouse_down 				= Event.MouseClick
	mouse_up   				= Event.MouseClick
	mouse_clicked			: Event.MouseClick
	building_print			: Event.Event
	char_received 			: Event.CharReceived
	selected_changed_manual : Event.Event
	print_build 		    : Event.Event
	after_print 		    : Event.Event
	focused_on_list_changed	: Event.FocusedOnListChanged


	def __init__(self,ctrl):
		self.mouse_clicked				= Event.MouseClick(ctrl)
		self.mouse_down 				= Event.MouseClick(ctrl)
		self.mouse_up   				= Event.MouseClick(ctrl)
		self.char_received 				= Event.CharReceived(ctrl)
		self.building_print				= Event.Event(ctrl)
		self.print_build				= Event.Event(ctrl)
		self.after_print				= Event.Event(ctrl)
		self.selected_changed_manual 	= Event.Event(ctrl)
		self.focused_on_list_changed	= Event.FocusedOnListChanged(ctrl)

		@self.mouse_up()
		def d(me:MouseClick):
			print("UP")
			print(me.Position)
			print("\n"*4)
		@self.mouse_down()
		def d(me:MouseClick):
			print("DOWN")
			print(me.Position)
			print("\n"*4)
	def init(self):
		for d in filter(lambda x : isinstance(getattr(self,x),Event), dir(self)):
			getattr(self,d).init()





class MouseState:
	pressed = False
	released = False
	bstate : int
	def __init__(self,bstate:int):
		self.bstate = bstate
		if bstate == 0:
			self.pressed = True
		if bstate == 1:
			self.released = True
		
	def __repr__(self):
		return f"p {self.pressed} r {self.released} s {self.bstate}"
	


#BUTTONn_PRESSED, BUTTONn_RELEASED, BUTTONn_CLICKED, BUTTONn_DOUBLE_CLICKED, BUTTONn_TRIPLE_CLICKED, BUTTON_SHIFT, BUTTON_CTRL, BUTTON_ALT




class LineConstructor:
	def __init__(self,text:str='',selected:bool=False,color:int=None,underline:bool=False,bold:bool=False,prefix:str='',add_line_break:bool=True,empty:int=0,x:int=None,y:int=None):
		self.text           = text
		self.selected       = selected
		self.color          = color
		self.underline      = underline
		self.bold           = bold
		self.prefix         = prefix
		self.add_line_break = add_line_break
		self.empty          = empty
		self.x              = x
		self.y              = y
	@property
	def attr(self):
		color = (33554432 if self.selected else 16777216) if self.color is None else self.color
		val  = color
		for should_do,attrValue in filter(lambda x : x[0],[[self.underline,curses.A_UNDERLINE],[self.bold,curses.A_BOLD]]):
			val = val | attrValue
		return val
	
	@property
	def line_height(self):
		if self.empty:
			return self.empty
		if self.add_line_break:
			return 1
		return 0
	
	def __repr__(self):
		return f'LineConstructor {self.prefix}{self.text}'
class Position:
	def __init__(self,x:int,y:int):
		self.x = x
		self.y = y
	def __repr__(self):
		return f'Position ({self.x},{self.y})'

	@classmethod
	def reversed(cls,y:int,x:int):
		return cls(x,y)
class Area:
	on_click:callable = None
	def pos_is_in(self,pos:Position) -> bool:
		return False
	def handle_click(self,pos:Position,*a,**b):
		if not self.on_click:
			return
		if self.pos_is_in(pos):
			self.on_click(*a,**b)


class LineArea(Area):
	def __init__(self,y:int,start_x:int,stop_x:int):
		self.y       = y
		self.start_x = start_x
		self.stop_x  = stop_x
	def pos_is_in(self,pos:Position) -> bool:
		return (pos.y == self.y) and (self.start_x <= pos.x <= self.stop_x)
	def __repr__(self):
		return f'LineArea [{self.y}][{self.start_x}-{self.stop_x}]'
class LinesArea(Area):
	lines : List[LineArea] = []
	def __init__(self):
		self.lines = []
	def append(self,line:LineArea):
		self.lines.append(line)
	def reset(self):
		self.lines = []
	def pos_is_in(self,pos:Position) -> bool:
		for l in self.lines:
			if l.pos_is_in(pos):
				return True
		return False
	def add_dummy(self,n=1):
		for i in range(n):
			self.append(LineArea(0,0,0))
	@property
	def line_range(self):
		if not self.lines:
			return [None,None]
		ar    = [l.y for l in self.lines]
		return [min(ar),max(ar)]

	def __repr__(self):
		sval = ','.join([str(x) for x in self.lines])
		return f'LinesArea {sval}]'
class InputChar:
	def __init__(self,code:int,name:str=None):
		self.code = code
		self.name = Curses.get_keyname(code) if name is None else name
		self.c    = chr(code)
	def __repr__(self):
		return f'Code : {self.code} c : {self.c} name : {self.name}'

	@classmethod
	def from_ch(cls,code:int=None) -> InputChar:
		return InputChar(code=code) if not (code is None or (code < 0)) else None
	
class Button(LinesArea):
	def __init__(self,text:str, win : curses.window,on_click=None,top=None,left=None,right=None,bottom=None):
		uly = 0
		ulx = 0
		lry = 0
		lrx = 0
		if top is not None:
			uly = top
			lry = uly + 2
		if bottom is not None:
			lry = bottom
			uly = lry - 2
		if left is not None:
			ulx = left
			lrx = ulx + len(text) + 2
		if right is not None:
			lrx = right
			ulx = lrx - len(text) - 2

		self.rec      = rectangle(win,uly,ulx,lry,lrx)
		self.text     = text
		self.on_click = on_click
		for y in range(uly,lry):
			self.append(LineArea(y,ulx,lrx))
		
		win.addstr(uly+1,ulx+1,text,LineConstructor(selected=True,bold=True,underline=True).attr)

class LineBlock(Area):
	lines : List[LineConstructor]
	on_click : callable
	linesDrawn : LinesArea
	item_index : int
	def __init__(self,*lines:List[LineConstructor],on_click:callable=None,on_click_ar=[],on_click_p=dict()):
		self.lines       = lines
		self.on_click    = on_click
		self.on_click_ar = on_click_ar
		self.on_click_p  = on_click_p
		self.linesDrawn  = LinesArea()
		self.item_index  = None

	def append(self,line:LineConstructor):
		self.lines.append(line)
	
	@property
	def line_height(self):
		return sum(l.line_height for l in self.lines)
	
	def pos_is_in(self,pos:Position) -> bool:
		return self.linesDrawn.pos_is_in(pos)
	
	def handle_mouse_clicked(self,pos:Position):
		if not self.on_click:
			return
		if self.pos_is_in(pos):
			self.on_click(*self.on_click_ar,**self.on_click_p)
	@property
	def line_range(self):
		return self.linesDrawn.line_range
	
class Curses:
	class _Color:
		normal   : int
		selected : int
		blue     : int
		mag      : int
		fullcyan : int
		def __init__(self,normal:int=0,selected:int=0):
			curses.start_color()
			def get(index:int,font:int,back:int) -> int:
				curses.init_pair(index,font,back)
				return curses.color_pair(index)
			self.normal		= get(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
			self.selected 	= get(2, curses.COLOR_BLACK, curses.COLOR_CYAN)
			self.blue		= get(3, curses.COLOR_WHITE, curses.COLOR_BLUE)
			self.mag		= get(4, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
			self.fullcyan	= get(4, curses.COLOR_CYAN, curses.COLOR_CYAN)

	color            : _Color             
	win              : curses.window       

	def __init__(self,max_lines:int=30):
		self.init()
		win : curses.window 	= curses.initscr()
		self.max_lines			= max_lines
		self.win 				= win
		self.color 				= Curses._Color()
		curses.mousemask(1)
		win.keypad(1)
		curses.mouseinterval(50)

	def clear(self):
		self.win.erase()
	
	def refresh(self):
		self.win.refresh()

	def stop(self):
		self.clear()
		self.refresh()
		curses.endwin()

	def draw_line_old(self,line:LineConstructor):
		if line.empty:
			self.win.addstr("\n"*line.empty)
			return
	
		color = (self.color.selected if line.selected else self.color.normal) if line.color is None else line.color
		attr  = color
		for should_do,attrValue in filter(lambda x : x[0],[[line.underline,curses.A_UNDERLINE],[line.bold,curses.A_BOLD]]):
			attr = attr | attrValue
		if line.prefix:
			self.win.addstr(line.prefix)
		self.win.addstr(line.text,attr)
		if line.add_line_break:
			self.win.addstr("\n")
	


	def parse_block(self,block:LineBlock):
		self.drawLines(block.lines,block.linesDrawn)
	
	def drawLines(self,lines:List[LineConstructor]=[],linesArea:LinesArea=None,debug=False) -> LinesArea:
		def draw_line(line:LineConstructor):
			color = (self.color.selected if line.selected else self.color.normal) if line.color is None else line.color
			attr  = color
			for should_do,attrValue in filter(lambda x : x[0],[[line.underline,curses.A_UNDERLINE],[line.bold,curses.A_BOLD]]):
				attr = attr | attrValue
			if line.prefix:
				self.win.addstr(line.prefix)
			if line.x is not None and line.y is not None:
				self.win.addstr(line.y,line.x,line.text,attr)
			else:
				self.win.addstr(line.text,attr)
		
	
		retval = LinesArea() if linesArea is None else linesArea
		for l in lines:
			
			if l.empty:
				self.win.addstr("\n"*l.empty)
			else:
				start_pos = self.cursor_position
				draw_line(l)
				
				end_pos = self.cursor_position
				
				retval.append(LineArea(start_pos.y,start_pos.x,end_pos.x))
				if l.add_line_break:
					self.win.addstr("\n")

		return retval

	@property
	def max_x(self):
		return self.win.getmaxyx()[1]
	@property
	def max_y(self):
		return self.win.getmaxyx()[0]
	@property
	def cursor_position(self) -> Position:
		return Position.reversed(*self.win.getyx())
	
	def input(self) -> InputChar:
		return InputChar.from_ch(self.win.getch())



	@classmethod
	def get_mouse(cls):
		_, mx, my, _, _ = curses.getmouse()
		return Position(mx,my)

	@classmethod
	def get_keyname(cls,keyname,quit=False):
		start_val = keyname
		if isinstance(keyname,int):
			if keyname in cls.codeToName:
				return cls.codeToName[keyname]
			keyname = chr(keyname)
		if keyname in cls.nameToCode:
			return keyname
		if len(keyname) == 1:
			code = ord(keyname)
			if code in cls.codeToName:
				return cls.codeToName[code]
		if quit:


			print(f'Invalid keyname {keyname} {start_val}')

			print(cls.codeToName)
			exit()
		return None
	
	__init_done = False


	@classmethod
	def init(cls):
		if cls.__init_done:
			return
		cls.__init_done = True
		cls.codeToName = {int(v) : k.lower()[4:] for k,v in curses.__dict__.items() if k.lower().startswith('key_')}
		extra_keys = dict(
			enter=[10,459],
			esc=[27],
			backspace=[8],
			tab=[9],
			mup=[],
			m1released=[1],
			m2pressed=[2]
		)
		for keyname,codes in extra_keys.items():
			for code in codes:
				cls.codeToName[code] = keyname
		cls.nameToCode = dict()
		for k,v in cls.codeToName.items():
			if not v in cls.nameToCode:
				cls.nameToCode[v] = []
			cls.nameToCode[v].append(k)

def find_index(list:List[MenuItem],index):
	for i in range(len(list)):
		if list[i].index == index:
			return i
	return -1


class MenuItem:
	index       : int       = -1
	menu        : Menu      = None
	text        : str       = ""
	action_func : callable  = None
	data	    : any		= None
	def __init__(self,text:str="text",menu:Menu=None,action_func:callable=None,data=None):
		self.text        = text
		self.menu        = menu
		self.action_func = action_func
		self.data		 = data
		self.menu.register_item(self)
	
	@property
	def is_selected(self):
		return self.menu.selected_item_index == self.index
	
	def action(self):
		if self.action_func:
			self.action_func(self)
	
	@property
	def filter_index(self):
		return find_index(self.menu.f_items,self.index)
	
	@property
	def display_index(self):
		return find_index(self.menu.d_items,self.index)
	
	@property
	def ctrl(self):
		return self.menu.ctrl
	
	def on_click(self):
		self.action()
	
	def write(self,terminalPart:TerminalPart,last_one:bool=False):
		l = len(str(len(self.menu.items)))
		prefix = f'{str(self.display_index).zfill(l)} .'
		prefix = f'{str(self.index).zfill(l)} {str(self.display_index).zfill(l)} .'
		prefix = f'{str(self.index).zfill(l)} {str(self.display_index).zfill(l)} {str(self.filter_index).zfill(l)} .  '
		add_line_break = not last_one
		


		terminalPart.create_block(LineConstructor(text=self.text,selected=self.is_selected,prefix=prefix,add_line_break=add_line_break),on_click=self.on_click)
	
	def __repr__(self):
		return f"[MenuItem] {self.index}"
	
class Menu:
	ctrl                : Ctrl           = None
	menu_index          : int            = -1
	items               : List[MenuItem] = []
	f_items             : List[MenuItem] = []
	d_items             : List[MenuItem] = []
	title               : str            = ""
	selected_item_index : int            = -1
	d_max_len			: int			 = 0
	__d_first_f_index	: int			 = 0
	not_manual		    : bool		 	 = False

	def __init__(self,title="Title",ctrl:Ctrl=None):
		self.title               = title
		self.ctrl                = ctrl
		self.d_max_len           = ctrl.cur.max_lines
		self.f_items             = []
		self.d_items             = []
		self.items               = []
		self.selected_item_index = 0
		self.ctrl.register_menu(self)
	def reset(self):
		self.f_items             = []
		self.d_items             = []
		self.items               = []
		self.selected_item_index = 0
		self.__d_first_f_index   = 0
	def items_sorted_by_text(self,items) -> List[MenuItem]:
		return list(sorted(items, key = lambda x : x.text.lower()))
	
	@property
	def d_first_f_index(self):
		return self.__d_first_f_index
	
	@d_first_f_index.setter
	def d_first_f_index(self,value):
		next_val = min(len(self.f_items) - self.d_max_len,value)
		next_val = max(0,next_val)
		self.__d_first_f_index = next_val
	
	@property
	def f_selected_index(self):
		return find_index(self.f_items,self.selected_item_index)
	
	@f_selected_index.setter
	def f_selected_index(self,f_new_index):
		gli           = good_list_index
		f_len         = len(self.f_items)
		f_last_index  = self.f_selected_index
		f_new_index   = gli(f_new_index,f_len)
		f_delta_index = f_new_index - f_last_index

		if not f_len:
			return
		
		if f_delta_index == 0:
			return
		
		last_sel_was_at_bot = self.d_items and self.d_items[-1] == self.selected_item_index
		last_sel_was_at_top = self.d_items and self.d_items[0]  == self.selected_item_index
		going_one_down 		= f_delta_index == 1
		going_one_up   		= f_delta_index == -1

		if going_one_down and last_sel_was_at_bot:
			self.d_first_f_index += 1
		elif going_one_up and last_sel_was_at_top:
			self.d_first_f_index -= 1
		else:
			self.d_first_f_index = f_new_index - math.floor(self.d_max_len/2)


		self.selected_item_index = self.f_items[f_new_index].index

		if self.not_manual:
			self.not_manual = False
		else:
			self.ctrl.event.selected_changed_manual.fire()

	def set_filter_selected_not_manual(self,index):
		self.not_manual       = True
		self.f_selected_index = index
		self.not_manual       = False
	
	@property
	def d_selected_index(self):
		return find_index(self.d_items,self.selected_item_index)
	
	def register_item(self,item:MenuItem):
		item.index = len(self.items)
		self.items.append(item)
	
	@property
	def selected_item(self):
		try:
			retval = self.items[self.selected_item_index]
			return retval
		except:
			return False
		
	def setActive(self):
		self.ctrl.active_menu_index = self.menu_index
	@property
	def active(self):
		return self.menu_index == self.ctrl.menu_active_index


	def print_info(self):
		print('----------\n')
		ar = []
		for k,v in dict(
			selected_item_index=self.selected_item_index,
			f_selected_index=self.f_selected_index,
			d_first_f_index=self.d_first_f_index,
			f_items_l=len(self.f_items),
			d_items_l=len(self.d_items)).items():
			ar.append([k,str(v)])
		for x in String.same_len_ar(ar):
			print(x)
	
	def set_item_f(self):
		self.f_items	= [item for item in self.items if all_true(self.ctrl.filters,item)]
	def set_item_d(self,max_lines=None):
		if max_lines is not None:
			self.d_max_len = max_lines
		self.d_items	= self.f_items[self.d_first_f_index:self.d_first_f_index+self.d_max_len]
	
	
	def item(self,text:str,data=None):
		def decor(f):
			item = MenuItem(text=text,menu=self,action_func=f,data=data)
			return f
		return decor
	
	def one_up(self):
		self.f_selected_index -= 1

	def one_down(self):
		self.f_selected_index += 1

	def page_up(self):
		self.f_selected_index -= self.d_max_len

	def page_down(self):
		self.f_selected_index += self.d_max_len
	



class TerminalPart:
	blocks : List[LineBlock]
	ctrl   : Ctrl

	def __init__(self,ctrl:Ctrl):
		self.ctrl   = ctrl
		self.blocks = []

		@ctrl.event.mouse_clicked()
		def on_click(ev:Event.MouseClick):
			for b in self.blocks:
				b.handle_mouse_clicked(ev.coord)

	def wrapper(self,*lines:List[LineConstructor],on_click_ar=[],on_click_p=dict()):
		def dec(on_click):
			self.append(LineBlock(*lines,on_click=on_click,on_click_ar=on_click_ar,on_click_p=on_click_p))
			return on_click
		return dec
	
	def create_block(self,*lines:List[LineConstructor],on_click:callable=None,on_click_ar=[],on_click_p=dict()):
		self.append(LineBlock(*lines,on_click=on_click,on_click_ar=on_click_ar,on_click_p=on_click_p))
	
	def append(self,block:LineBlock):
		self.blocks.append(block)

	@property
	def line_height(self):
		return sum(b.line_height for b in self.blocks)

	@property
	def line_range(self):
		ar = []
		for block in self.blocks:
			val = block.line_range
			if val is not None:
				ar += val


		if not ar:
			return [None,None]
		return [min(ar),max(ar)]
	
	def write(self,c:Curses):
		for block in self.blocks:
			c.parse_block(block)

class ItemTerminalPart(TerminalPart):
	def add_item_index(self,block_index_start:int,item_index:int):
		if len(self.blocks) > block_index_start:
			for i in range(block_index_start,len(self.blocks)):
				self.blocks[i].item_index = item_index
	


	def fix_max_height(self,max_height:int):
		def delete_last_item():
			last_item_index = self.blocks[-1].item_index
			start_len = len(self.blocks)
			while self.blocks and self.blocks[-1].item_index == last_item_index:
				self.blocks = self.blocks[:-1]
		

		if max_height < 0:
			print(f"ItemTerminalPart fix_max_height max_height < 0 {max_height} ")
			exit()
		while self.blocks and (self.line_height > max_height):
			delete_last_item()
	

class Ctrl:

	#waiting_for_input : bool 			 = False
	menus             : List[Menu]       = []
	cur               : Curses           = None
	active_menu_index : int              = -1
	__on_key          : dict             = None
	event             : Events	 		 = None
	__focused_on_list : bool             = True
	running           : bool             = True
	terminal_top      : TerminalPart     = None
	terminal_bot      : TerminalPart     = None
	terminal_items    : ItemTerminalPart = None
	filters			  : List[callable]   = []
	def __init__(self,max_lines=30):
		self.filters 		   = []
		self.cur               = Curses(max_lines=max_lines)
		self.menus             = []
		self.__on_key          = dict()
		self.event             = Events(self)
		self.__focused_on_list = True
		self.running		   = False
		self.terminal_top	   = TerminalPart(self)
		self.terminal_bot	   = TerminalPart(self)
		self.terminal_items	   = ItemTerminalPart(self)
		self.__register_event_listerners()
	@property
	def active_menu(self):
		try:
			retval = self.menus[self.active_menu_index]
			return retval
		except:
			return False

	@property
	def selected_item(self):
		try:
			retval = self.active_menu.selected_item
			return retval
		except:
			return False
	
	def create_menu(self,title:str):
		return Menu(title=title,ctrl=self)
	
	def register_menu(self,menu:Menu):
		menu.menu_index = len(self.menus)
		if self.active_menu_index == -1:
			self.active_menu_index = menu.menu_index
		
		self.menus.append(menu)

	
	@property
	def focused_on_list(self):
		return self.__focused_on_list
	
	@focused_on_list.setter
	def focused_on_list(self,value):
		if value != self.focused_on_list:
			self.__focused_on_list = value
			self.event.focused_on_list_changed.fire(value)

	def __register_event_listerners(self):
		key = self.on_key

		def condition_active_menu_exists(*a,**b):
			return True if self.active_menu else False
		
		@key('esc')
		def f(ctrl:Ctrl):
			ctrl.stop()
		
		@key('up',condition=condition_active_menu_exists,no_para=True)
		def f():	self.active_menu.one_up()
		
		@key('down',condition=condition_active_menu_exists,no_para=True)
		def f():	self.active_menu.one_down()
		
		@key('enter',no_error=True)
		def f(ctrl:Ctrl):				ctrl.selected_item.action()
		
		@key('ppage',condition=condition_active_menu_exists)
		def f(ctrl:Ctrl):				ctrl.active_menu.page_up()
		
		@key('npage',condition=condition_active_menu_exists)
		def f(ctrl:Ctrl): 				ctrl.active_menu.page_down()



		@key(curses.BUTTON1_RELEASED)
		def button_up(ctrl:Ctrl):	
			ctrl.event.mouse_up.fire()


		@key(curses.BUTTON1_PRESSED)
		def mouse_down(ctrl:Ctrl):
			print("pressed")
			ctrl.event.mouse_down.fire()

		@key(curses.KEY_MOUSE)
		def mouse_clicked(ctrl:Ctrl):	
			ctrl.event.mouse_clicked.fire()


		@key(curses.KEY_F10,no_para=True)
		def ok():
			self.active_menu_index = -1
			print(f"self.active_menu_index {self.active_menu_index}")
		@self.event.building_print(condition=condition_active_menu_exists)
		def check_title(ev:Event.Event):
			m = self.active_menu
			if not m.title:
				return
			self.terminal_top.create_block(LineConstructor(text=m.title,underline=True),LineConstructor(empty=1))
	
	def on_key(self,key,func=None,condition=None,no_error=False,no_para=False):

		keyname = Curses.get_keyname(key,quit=True)
		if func is not None:
			self.__on_key[keyname] = convert_func(func,no_error=no_error,condition=condition,no_para=no_para)
			return
		
		def decorator(f):
			ret_func = convert_func(f,no_error=no_error,condition=condition,no_para=no_para)
			self.__on_key[keyname] = ret_func
			return ret_func
		return decorator
	
	def do_key(self,code):
		keyname = Curses.get_keyname(code)
		if not keyname:
			return
		if keyname in self.__on_key:
			self.__on_key[keyname](self)
	
	def exit(self):
		self.stop()
		self.cur.stop()
	def stop(self):
		self.running = False
	
	def display(self):
		self.running = True
		def printer(c:Curses,m:Menu):
			for term in [self.terminal_top,self.terminal_items,self.terminal_bot]:
				term.blocks = []
			c.clear()
			self.cur.max_lines = c.win.getmaxyx()[0] - 1
			m.set_item_f()
			self.event.building_print.fire()
			m.set_item_f()
			m.set_item_d(max_lines=self.cur.max_lines - self.terminal_top.line_height - self.terminal_bot.line_height)
			term_item = self.terminal_items

			d_items_len = len(m.d_items)
			d_items_last_index = len(m.d_items) - 1
			for list_index in range(d_items_len):
				item : MenuItem = m.d_items[list_index]
				last_len 		= len(term_item.blocks)
				item.write(term_item,last_one=list_index==d_items_last_index)
				term_item.add_item_index(last_len,item.index)

			term_item.fix_max_height(self.cur.max_lines - self.terminal_top.line_height - self.terminal_bot.line_height)
			self.event.print_build.fire()


			for term in [self.terminal_top,self.terminal_items,self.terminal_bot]:
				term.write(c)
			
			self.event.after_print.fire()
			inputChar = c.input()
			self.event.char_received.fire(inputChar)
			self.do_key(inputChar.code)

		while self.running and self.active_menu:
			printer(self.cur,self.active_menu)
		self.cur.stop()

	def filter(self,f):
		self.filters.append(f)
		return f


from . import ValueGetter
from ..models import InputParameter,copy_tab,IHasReplacement_tab
from .. import Validator
from typing import List
from jivi.util.ob import get_value_stack_from_supers

class UIInput(IHasReplacement_tab):
	para            : InputParameter            = None
	_method_classes : List[callable]            = []
	_validators     : List[Validator.Validator] = []

	def __init__(self,para : InputParameter):
		self.para = para
	@property
	def my_replacement_tab(self):
		return dict()
	
	@property
	def replacement_tab(self):
		tab = copy_tab(self.para.replacement_tab)
		tab.update(self.my_replacement_tab)
		return tab
	
	@property
	def inputMethods(self):
		return [cls(self) for cls in self._method_classes]
	
	def __repr__(self):
		return f"[<{self.__class__.__name__}>] {self.para}"

	@property
	def validators(self) -> List[Validator.Validator]:
		return [val(self) for val in get_value_stack_from_supers(self,'_validators',True)]

	
class File(UIInput):
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter]

	@property
	def validators(self) -> List[Validator.Validator]:

		return []
	
class Json(File):
	validators      = [Validator.FileExists]
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter,ValueGetter.Text]

class Text(UIInput):
	_method_classes = [ValueGetter.Filepath,ValueGetter.FileSelecter,ValueGetter.Text]

class String(UIInput):
	_method_classes = [ValueGetter.String]

class Number(UIInput):
	_method_classes = [ValueGetter.Number]


from typing import TypeVar, Generic,List,Dict
from typing import Callable

from logging import Logger
from ..callback import *


T = TypeVar('T')


class Event:
	__callbacks : List[CB]
	_is_single  : bool     = False
	_running    : bool     = False
	def __init__(self,single=False,before_fire=None):
		self._is_single  = single
		self.__callbacks = []
		if before_fire:
			self.before_fire = before_fire
		
	def before_fire(self,*a,**b):
		pass

	def fire(self,*a,**b):
		self._running = True
		self.before_fire(*a,**b)
		for cb in sorted(self.__callbacks,key=lambda x : x.options.priority):
			if self._running:
				cb(self)
			else:
				break
	
	def stopPropagation(self):
		self._running = False
	
	@property
	def cb(self):
		return CBOptions().set
	
	def register(self,options:CBOptions=CBOptions(),func=None):
		def add_func(f: Callable[[self.__class__],None]) -> Callable[[self.__class__],None]:
			cb = CB(f,options)
			if self._is_single:
				self.__callbacks = [cb]
			else:
				self.__callbacks.append(cb)
			retval : Callable[[self.__class__],None] = f
			return retval
		
		if func:
			return add_func(func)
		
		return add_func
	def __call__(self,options:CBOptions=CBOptions(),func=None):
		return self.register(options,func=func)

class EventWithParent(Event,Generic[T]):
	parent   : T
	def __init__(self,single=False,before_fire=None,parent:T=None):
		super().__init__(single=single,before_fire=before_fire)
		self.parent = parent


class EventWithKey:
	__events     : Dict[str,Event] = dict()
	eventClass   : Event           = Event
	_is_single   : bool            = True
	_before_fire : callable        = None
	def __init__(self,single=False,eventClass=Event,format_key=None,valid_key=None,before_fire=None):
		self.__events   = dict()
		self.eventClass = eventClass
		self._is_single = single
		if format_key:
			self.format_key = format_key
		if valid_key:
			self.valid_key = valid_key
		
		self._before_fire = before_fire

	@property
	def cb(self):
		return CBOptions().set
	
	def format_key(self,inp):
		return inp
	
	def valid_key(self,key):
		return True

	def _create(self):
		return self.eventClass(single=self._is_single,before_fire=self._before_fire)

	def get(self,key):
		key = self.format_key(key)
		if key in self.__events:
			return self.__events[key]
		if self.valid_key(key):
			self.__events[key] = self._create()
			return self.__events[key]

	def fire(self,key,*a,**b):
		x = self.get(key)
		if not x:
			return
		x.fire(*a,**b)

	def register(self,key,options:CBOptions=CBOptions(),func=None):
		if ev := self.get(key):
			return ev.register(options,func=func)

class EventWithKeyWithParent(EventWithKey,Generic[T]):
	parent   : T
	__events     : Dict[str,EventWithParent[T]] = dict()

	def __init__(self,single=False,eventClass=EventWithParent[T],format_key=None,valid_key=None,before_fire=None,parent:T=None):
		super().__init__(single=single,eventClass=eventClass,format_key=format_key,valid_key=valid_key,before_fire=before_fire)
		self.parent = parent

	def _create(self):
		return self.eventClass(single=self._is_single,before_fire=self._before_fire,parent=self.parent)
	


from __future__ 			import annotations
import curses
from ..util import random_el,convert_func

class COLOR:
	class B:
		BLACK   = 0
		BLUE    = 0
		CYAN    = 0
		GREEN   = 0
		MAGENTA = 0
		RED     = 0
		WHITE   = 0
		YELLOW  = 0
		ALL     = []
		@classmethod
		def init(cls):
			bad_keys = ["init","ALL","random"]
			for a  in dir(cls):
				if a.startswith('_') or a in bad_keys:
					continue
				val = getattr(curses,f"COLOR_{a}")
				retval = CURSES.create_color(curses.COLOR_BLACK,val)
			
				setattr(cls,a,retval)
				if a != "BLACK":
					cls.ALL.append(retval)
		@classmethod
		def random(cls):
			return random_el(cls.ALL)

	class F:
		BLACK   = 0
		BLUE    = 0
		CYAN    = 0
		GREEN   = 0
		MAGENTA = 0
		RED     = 0
		WHITE   = 0
		YELLOW  = 0
		ALL     = []
		@classmethod
		def init(cls):
			bad_keys = ["init","ALL","random"]
			for a  in dir(cls):
				if a.startswith('_') or a in bad_keys:
					continue
				retval = CURSES.create_color(getattr(curses,f"COLOR_{a}"),curses.COLOR_BLACK)
				setattr(cls,a,retval)
				if a != "BLACK":
					cls.ALL.append(retval)
		
		@classmethod
		def random(cls):
			return random_el(cls.ALL)

	@classmethod
	def init(cls):
		cls.B.init()
		cls.F.init()
		

def _curses_get_keys():
	codeToName = {int(v) : k.lower()[4:] for k,v in curses.__dict__.items() if k.lower().startswith('key_')}
	extra_keys = dict(
		enter=[10,459],
		esc=[27],
		space=[32],
		backspace=[8],
		tab=[9],
		mup=[],
		m1released=[1],
		m2pressed=[2]
	)
	for keyname,codes in extra_keys.items():
		for code in codes:
			codeToName[code] = keyname
	nameToCode = dict()
	for k,v in codeToName.items():
		if not v in nameToCode:
			nameToCode[v] = []
		nameToCode[v].append(k)
	return codeToName,nameToCode

class CURSES:
	codeToName,nameToCode = _curses_get_keys()

	@classmethod
	def get_keyname(cls,keyname,quit=False):
		def failed():
			if quit:


				print(f'Invalid keyname {keyname} {start_val}')

				print(cls.codeToName)
				exit()
			return None

		start_val = keyname
		if isinstance(keyname,int):
			if keyname in cls.codeToName:
				return cls.codeToName[keyname]
			keyname = chr(keyname)
		if keyname in cls.nameToCode:
			return keyname
		if not keyname:
			return failed()
		if len(keyname) == 1:
			code = ord(keyname)
			if code in cls.codeToName:
				return cls.codeToName[code]

		return failed()

	__last_color_index = 1
	@classmethod
	def create_color(cls,font:int,back:int) -> int:
		curses.init_pair(cls.__last_color_index,font,back)
		retval =  curses.color_pair(cls.__last_color_index)
		cls.__last_color_index += 1
		return retval

	@classmethod
	def init(cls):
		COLOR.B.init()
		COLOR.F.init()
	
class Position:
	def __init__(self,x:int,y:int):
		self.x = x
		self.y = y
	def __repr__(self):
		return f'Position ({self.x},{self.y})'

	
	@classmethod
	def reversed(cls,y:int,x:int):
		return cls(x,y)

class InputChar:
	def __init__(self,code:int,name:str=None):
		self.code = code
		self.name = CURSES.get_keyname(code) if name is None else name
		self.c    = chr(code)
	def __repr__(self):
		return f'Code : {self.code} c : {self.c} name : {self.name}'

	@classmethod
	def from_ch(cls,code:int=None) -> InputChar:
		return InputChar(code=code) if not (code is None or (code < 0)) else None

class OnKeyCb:
	def __init__(self,func:callable,args=[],kwargs=dict(),condition=None,no_error=False):
		self.func      = func
		self.condition = condition
		self.no_error  = no_error
		self.args         = args
		self.kwargs         = kwargs


	def do_func(self):
		if self.condition and not self.condition(*self.args,**self.kwargs):
			return
		
		return self.func(*self.args,**self.kwargs)

	
	def call(self):
		if self.no_error:
			try:
				retval = self.do_func()
				return retval
			except:
				pass
		else:
			return self.do_func()



class CursesWrapper:
	win:curses.window
	running:bool=False
	def __init__(self):
		self.__on_key = dict()
		

	def initscr(self,win:curses.window):
		self.win = win
		CURSES.init()

	def on_key_old(self,key,func=None,condition=None,no_error=False,no_para=False,no_main=False):

		# convert_func_options = dict(
		# 	no_error=no_error,
		# 	condition=condition,
		# 	no_para=no_para,
		# 	kwargs=kwargs,
		# 	args=([self] if not no_main else []) + args)

		convert_func_options = dict(
			no_error=no_error,
			condition=condition,
			no_para=no_para)

		keyname = CURSES.get_keyname(key,quit=True)
		if func is not None:
			self.__on_key[keyname] = convert_func(func,**convert_func_options)
			return
		
		def decorator(f):
			ret_func = convert_func(f,**convert_func_options)
			self.__on_key[keyname] = ret_func
			return ret_func
		return decorator

	def on_key(self,key,func=None,condition=None,no_error=False,no_arg_default=False,args=[],kwargs=dict()):
		keyname = CURSES.get_keyname(key,quit=True)
		construct_dict = dict(
			condition=condition,
			no_error=no_error,
			args=args if no_arg_default else [self] + args,
			kwargs=kwargs)
		if func is not None:
			self.__on_key[keyname] = OnKeyCb(func,**construct_dict)
			return
		
		def decorator(f):
			self.__on_key[keyname] = OnKeyCb(f,**construct_dict)
			return f
		return decorator

	def do_key_old(self,code):

		keyname = CURSES.get_keyname(code)
		
		if not keyname:
			return
		if keyname in self.__on_key:
			self.__on_key[keyname](self)
	
	def do_key(self,code):

		keyname = CURSES.get_keyname(code)
		
		if not keyname:
			return
		if keyname in self.__on_key:
			self.__on_key[keyname].call()

	

	def clear(self):
		self.win.erase()
	
	def refresh(self):
		self.win.refresh()

	def stop(self):
		self.clear()
		self.refresh()
		curses.endwin()
		self.running = False

	def key_reader(self):
		self.running = True
		while self.running:
			inputChar = self.input()
			self.do_key(inputChar.code)
	
	def input(self) -> InputChar:
		return InputChar.from_ch(self.win.getch())



def valid_key(inp):
	return True if CURSES.get_keyname(inp) else False


def get_keyname(inp):
	return CURSES.get_keyname(inp)


from typing import List

class CBOptions:
	clear_args : bool     	= False
	add_args   : List[any]	= []
	add_kwargs : dict     	= dict()
	condition  : callable 	= None
	no_error   : bool     	= False
	priority   : int		= 2
	def __init__(self,
				priority 			   = 2,
				clear_args             = False,
				add_args               = [],
				add_kwargs             = dict(),
				condition              = None,
				no_error			   = False):
		self.set(priority=priority,clear_args=clear_args,add_args=add_args,add_kwargs=add_kwargs,condition=condition,no_error=no_error)
	
	def set(self,priority=priority,clear_args= False,add_args= [],add_kwargs= dict(),condition=None,no_error=False):
		self.priority   = priority
		self.clear_args = clear_args
		self.add_args   = add_args
		self.add_kwargs = add_kwargs
		self.condition  = condition
		self.no_error   = no_error
		return self
	
	@property
	def copy(self):
		return self.__class__(priority=self.priority,clear_args=self.clear_args,add_args=self.add_args,add_kwargs=self.add_kwargs,condition=self.condition,no_error=self.no_error)


class CB:
	def __init__(self,func : callable,options:CBOptions=CBOptions()):
		self.original_func = func
		self.func          = self.original_func
		self.options       = options
		if options.condition:
			self.__convert_condition(options.condition)
		if options.no_error:
			self.__convert_no_error()
	
	def __convert_condition(self,condition):
		f = self.func
		def new_func(*a,**b):
			if condition(*a,**b):
				return f(*a,**b)
		self.func = new_func

	def __convert_no_error(self):
		f = self.func
		def new_func(*a,**b):
			try:
				retval = f(*a,**b)
				return retval
			except:
				pass
		self.func = new_func
	
	def __call__(self,*a,**b):
		if self.options.clear_args:
			return self.func(*self.options.add_args,**self.options.add_kwargs)
		else:
			for k,v in self.options.add_kwargs.items():
				b[k] = v
			return self.func(*a,*self.options.add_args,**b)


from ..wrapper 	import jvWrapper,jvWrapperSuper
from typing 	import List,Dict
from ..fs		import JDir,JFile
from ..uinput	import UInput


inputType = UInput.type

# class ConfigFieldType:
# 	inputType = InputType.string
# 	@classmethod
# 	def is_valid(cls,val):
# 		return isinstance(val,str)

# class ConfigFieldTypeString(ConfigFieldType):

# 	@classmethod
# 	def is_valid(cls,val):
# 		return isinstance(val,str)

# class ConfigFieldTypeFile(ConfigFieldTypeString):
# 	inputType = InputType.file
# 	@classmethod
# 	def is_valid(cls,val):
# 		return super().is_valid(val) and JFile(val).exists

# class ConfigFieldTypeJsonFile(ConfigFieldTypeString):
# 	inputType = InputType.json
# 	@classmethod
# 	def is_valid(cls,val):
# 		return super().is_valid(val) and JFile(val).exists



class ConfigField:
	key 		: str
	inputType 	: UInput.type.Type
	__value     : any

	def __init__(self,inputType,default=None):
		self.__value    = None
		self.inputType  = inputType
		self.default 	= default
	
	def __repr__(self):
		return f"[ConfigField] {self.key}"
	
	def set_value_if_not_set(self,val):
		if self.is_set:
			return
		self.value = val
	
	def try_setting(self):
		val = self.inputType.get_input(self.key)
		
	
	@property
	def value(self):
		if not self.is_set:
			return self.default
		
		return self.__value

	@value.setter
	def value(self,val):
		if self.inputType.is_valid(val):
			self.__value = val
			return
		print(f"Not valid value {val}")
	
	@property
	def is_set(self):
		return self.__value is not None
	

default_classifiers = ["Programming Language :: Python :: 3","License :: OSI Approved :: MIT License","Operating System :: OS Independent"]

@jvWrapper()
class ConfigCreater(jvWrapperSuper):
	name                          						= ConfigField(inputType.String)
	author                        						= ConfigField(inputType.String)
	author_email                  						= ConfigField(inputType.String)
	description                   						= ConfigField(inputType.String)
	long_description              						= ConfigField(inputType.File)
	long_description_content_type 						= ConfigField(inputType.String,default="text/markdown")
	url                           						= ConfigField(inputType.String)
	classifiers 				  						= ConfigField(inputType.JsonFile,default=default_classifiers)

	__all_fields 										= dict()
	def __init__(self,*json_config_files:List[str]):
		self.__all_fields = dict()
		for k,v in map(lambda x : (x,getattr(self,x)),filter(lambda k : isinstance(getattr(self,k),ConfigField),dir(self))):
			self.__all_fields[k] = v
			v.key = k
		
		self.read_from_file(*json_config_files)
		self.set_unset()

	def set_unset(self):
		for k,v in self.fields_not_set.items():
			v.try_setting()
			
	
	@property
	def fields_not_set(self) -> Dict[str,ConfigField]:
		return {k : v for k,v in self.__all_fields.items() if not v.is_set}
			


	
	def read_from_file(self,*filenames:List[str]):
		for filename in filenames:
			jFile = JFile(filename)
			if not jFile.exists:
				self.jvlog.warning(f"Filename {filename} not found!")
				continue
			jTab = jFile.reader.json()
			if not jTab:
				self.jvlog.warning(f"Filename {filename} unable to read json!")
				continue
			for k,v in self.__all_fields.items():
				if k in jTab:
					v.set_value_if_not_set(jTab[k])
			

